@@ -17,6 +17,7 @@
 #undef BALSA_REVISION
 #undef BALSA_VERSION
 #undef BALSA_SHOW_ALL
+#undef BALSA_MDN_REPLY
 #undef PACKAGE
 #undef VERSION
 
@@ -27,13 +28,20 @@
 #undef SENDMAIL
 #undef USE_POP
 #undef USE_IMAP
+#undef USE_SASL
+#undef USE_SSL
 
 /* Shall we multithread? */
 #undef BALSA_USE_THREADS
+#undef USE_PIXBUF
 
 /* LDAP support? */
 #undef ENABLE_LDAP
 
+/* ESMTP support? */
+#undef ENABLE_ESMTP
+#undef HAVE_SMTP_TLS_CLIENT_CERTIFICATE
+
 #undef BALSA_STD_PREFIX
 #undef BALSA_DATA_PREFIX
 #undef GNOME_STD_PREFIX
@@ -42,3 +50,6 @@
 
 /* The prefixes to look for balsa data, at least most likely */
 #undef BALSA_COMMON_PREFIXES
+
+/* use pcre instead of `standard' rexexp's? */
+#undef HAVE_PCRE
