@@ -29,39 +29,58 @@
 #include <gnome.h>
 
 #include <stdio.h>
+#include <sys/stat.h>
 
 #include "address-book.h"
 #include "address-book-vcard.h"
 #include "information.h"
 
+/* FIXME: Perhaps the whole thing could be rewritten to use a g_scanner ?? */
+
 /* FIXME: Arbitrary constant */
-/* Perhaps the whole thing could be rewritten to use a g_scanner ?? */
 #define LINE_LEN 256
+/* FIXME: Make an option */
+#define CASE_INSENSITIVE_NAME
 
 static GtkObjectClass *parent_class = NULL;
 
-static void
-libbalsa_address_book_vcard_class_init(LibBalsaAddressBookVcardClass *
-				       klass);
-static void libbalsa_address_book_vcard_init(LibBalsaAddressBookVcard *
-					     ab);
+typedef struct _CompletionData CompletionData;
+struct _CompletionData {
+    gchar *string;
+    LibBalsaAddress *address;
+};
+
+static void libbalsa_address_book_vcard_class_init(LibBalsaAddressBookVcardClass *klass);
+static void libbalsa_address_book_vcard_init(LibBalsaAddressBookVcard *ab);
 static void libbalsa_address_book_vcard_destroy(GtkObject * object);
 
-static void libbalsa_address_book_vcard_load(LibBalsaAddressBook * ab);
-static void libbalsa_address_book_vcard_store_address(LibBalsaAddressBook *
-						      ab,
-						      LibBalsaAddress *
-						      new_address);
+static void libbalsa_address_book_vcard_load(LibBalsaAddressBook * ab, 
+					     LibBalsaAddressBookLoadFunc callback, 
+					     gpointer closure);
+static void libbalsa_address_book_vcard_store_address(LibBalsaAddressBook *ab,
+						      LibBalsaAddress *new_address);
 
-static void libbalsa_address_book_vcard_save_config(LibBalsaAddressBook *
-						    ab,
+static void libbalsa_address_book_vcard_save_config(LibBalsaAddressBook *ab,
 						    const gchar * prefix);
-static void libbalsa_address_book_vcard_load_config(LibBalsaAddressBook *
-						    ab,
+static void libbalsa_address_book_vcard_load_config(LibBalsaAddressBook *ab,
 						    const gchar * prefix);
+static GList *libbalsa_address_book_vcard_alias_complete(LibBalsaAddressBook * ab,
+							 const gchar * prefix,
+							 gchar ** new_prefix);
 
 static gchar *extract_name(const gchar * string);
 
+static CompletionData *completion_data_new(LibBalsaAddress * address,
+					   gboolean alias);
+static void completion_data_free(CompletionData * data);
+static gchar *completion_data_extract(CompletionData * data);
+static gint address_compare(LibBalsaAddress *a, LibBalsaAddress *b);
+
+static void load_vcard_file(LibBalsaAddressBook *ab);
+
+static gboolean vcard_address_book_need_reload(LibBalsaAddressBookVcard *ab);
+
+
 GtkType libbalsa_address_book_vcard_get_type(void)
 {
     static GtkType address_book_vcard_type = 0;
@@ -109,12 +128,21 @@ libbalsa_address_book_vcard_class_init(LibBalsaAddressBookVcardClass *
 	libbalsa_address_book_vcard_save_config;
     address_book_class->load_config =
 	libbalsa_address_book_vcard_load_config;
+
+    address_book_class->alias_complete =
+	libbalsa_address_book_vcard_alias_complete;
+
 }
 
 static void
 libbalsa_address_book_vcard_init(LibBalsaAddressBookVcard * ab)
 {
     ab->path = NULL;
+    ab->address_list = NULL;
+    ab->mtime = 0;
+
+    ab->name_complete  = g_completion_new((GCompletionFunc)completion_data_extract);
+    ab->alias_complete = g_completion_new((GCompletionFunc)completion_data_extract);
 }
 
 static void
@@ -126,6 +154,16 @@ libbalsa_address_book_vcard_destroy(GtkObject * object)
 
     g_free(addr_vcard->path);
 
+    g_list_foreach(addr_vcard->address_list, (GFunc) gtk_object_unref, NULL);
+    g_list_free(addr_vcard->address_list);
+    addr_vcard->address_list = NULL;
+
+    g_list_foreach(addr_vcard->name_complete->items, (GFunc)completion_data_free, NULL);
+    g_list_foreach(addr_vcard->alias_complete->items, (GFunc)completion_data_free, NULL);
+    
+    g_completion_free(addr_vcard->name_complete);
+    g_completion_free(addr_vcard->alias_complete);
+
     if (GTK_OBJECT_CLASS(parent_class)->destroy)
 	(*GTK_OBJECT_CLASS(parent_class)->destroy) (GTK_OBJECT(object));
 
@@ -146,23 +184,67 @@ libbalsa_address_book_vcard_new(const gchar * name, const gchar * path)
     return ab;
 }
 
+static gboolean 
+vcard_address_book_need_reload(LibBalsaAddressBookVcard *ab)
+{
+    struct stat stat_buf;
+
+    if ( stat(ab->path, &stat_buf) == -1 ) {
+	libbalsa_information(LIBBALSA_INFORMATION_WARNING, _("Could not stat vcard address book: %s"), ab->path);
+	return FALSE;
+    }
+    if ( stat_buf.st_mtime > ab->mtime ) {
+	ab->mtime = stat_buf.st_mtime;
+	return TRUE;
+    } else {
+	return FALSE;
+    }
+}
+
 static void
-libbalsa_address_book_vcard_load(LibBalsaAddressBook * ab)
+libbalsa_address_book_vcard_load(LibBalsaAddressBook * ab, LibBalsaAddressBookLoadFunc callback, gpointer closure)
+{
+    GList *lst;
+
+    load_vcard_file(ab);
+
+    lst = LIBBALSA_ADDRESS_BOOK_VCARD(ab)->address_list;
+    while (lst) {
+	if ( callback ) 
+	    callback(ab, LIBBALSA_ADDRESS(lst->data), closure);
+	lst = g_list_next(lst);
+    }
+    callback(ab, NULL, closure);
+}
+
+/* FIXME: Could stat the file to see if it has changed since last time we read it */
+static void load_vcard_file(LibBalsaAddressBook *ab)
 {
     FILE *gc;
     gchar string[LINE_LEN];
     gchar *name = NULL, *id = NULL;
     gint in_vcard = FALSE;
     GList *list = NULL;
+    GList *completion_list = NULL;
     GList *address_list = NULL;
+    CompletionData *cmp_data;
 
     LibBalsaAddressBookVcard *addr_vcard;
 
     addr_vcard = LIBBALSA_ADDRESS_BOOK_VCARD(ab);
 
-    g_list_foreach(ab->address_list, (GFunc) gtk_object_unref, NULL);
-    g_list_free(ab->address_list);
-    ab->address_list = NULL;
+    if ( vcard_address_book_need_reload(addr_vcard) == FALSE ) 
+	return;
+
+    g_list_foreach(addr_vcard->address_list, (GFunc) gtk_object_unref, NULL);
+    g_list_free(addr_vcard->address_list);
+    addr_vcard->address_list = NULL;
+
+    g_list_foreach(addr_vcard->name_complete->items, (GFunc)completion_data_free, NULL);
+    g_list_foreach(addr_vcard->alias_complete->items, (GFunc)completion_data_free, NULL);
+
+    g_completion_clear_items(addr_vcard->name_complete);
+    g_completion_clear_items(addr_vcard->alias_complete);
 
     gc = fopen(addr_vcard->path, "r");
 
@@ -201,9 +283,9 @@ libbalsa_address_book_vcard_load(LibBalsaAddressBook * ab)
 		    address->full_name = g_strdup(_("No-Name"));
 
 
-		/* FIXME: Split into Firstname and Lastname... */
+		/* FIXME: Split into Firstname, Middlename and Lastname... */
 
-		list = g_list_append(list, address);
+		list = g_list_prepend(list, address);
 		address_list = NULL;
 	    } else {		/* record without e-mail address, ignore */
 		g_free(name);
@@ -243,7 +325,29 @@ libbalsa_address_book_vcard_load(LibBalsaAddressBook * ab)
     }
     fclose(gc);
 
-    ab->address_list = list;
+    list = g_list_sort(list, (GCompareFunc)address_compare);
+    addr_vcard->address_list = list;
+
+    completion_list = NULL;
+    while ( list ) {
+	cmp_data = completion_data_new(LIBBALSA_ADDRESS(list->data), FALSE);
+	completion_list = g_list_prepend(completion_list, cmp_data);
+	list = g_list_next(list);
+    }
+    completion_list = g_list_reverse(completion_list);
+    g_completion_add_items(addr_vcard->name_complete, completion_list);
+    g_list_free(completion_list);
+
+    completion_list = NULL;
+    list = addr_vcard->address_list;
+    while( list ) {
+	cmp_data = completion_data_new(LIBBALSA_ADDRESS(list->data), TRUE);
+	completion_list = g_list_prepend(completion_list, cmp_data);
+	list = g_list_next(list);
+    }
+    completion_list = g_list_reverse(completion_list);
+    g_completion_add_items(addr_vcard->alias_complete, completion_list);
+    g_list_free(completion_list);
 }
 
 static gchar *
@@ -269,19 +373,19 @@ extract_name(const gchar * string)
     name_arr = g_malloc((cpt + 1) * sizeof(gchar *));
 
     j = 0;
-    if (cpt > PREFIX && strlen(fld[PREFIX]) != 0)
+    if (cpt > PREFIX && *fld[PREFIX] != '\0')
 	name_arr[j++] = g_strdup(fld[PREFIX]);
 
-    if (cpt > FIRST && strlen(fld[FIRST]) != 0)
+    if (cpt > FIRST && *fld[FIRST] != '\0')
 	name_arr[j++] = g_strdup(fld[FIRST]);
 
-    if (cpt > MIDDLE && strlen(fld[MIDDLE]) != 0)
+    if (cpt > MIDDLE && *fld[MIDDLE] != '\0')
 	name_arr[j++] = g_strdup(fld[MIDDLE]);
 
-    if (cpt > LAST && strlen(fld[LAST]) != 0)
+    if (cpt > LAST && *fld[LAST] != '\0')
 	name_arr[j++] = g_strdup(fld[LAST]);
 
-    if (cpt > SUFFIX && strlen(fld[SUFFIX]) != 0)
+    if (cpt > SUFFIX && *fld[SUFFIX] != '\0')
 	name_arr[j++] = g_strdup(fld[SUFFIX]);
 
     name_arr[j] = NULL;
@@ -303,13 +407,12 @@ libbalsa_address_book_vcard_store_address(LibBalsaAddressBook * ab,
 					  LibBalsaAddress * new_address)
 {
     GList *list;
-    gchar *output;
     LibBalsaAddress *address;
     FILE *fp;
 
-    libbalsa_address_book_load(ab);
+    load_vcard_file(ab);
 
-    list = ab->address_list;
+    list = LIBBALSA_ADDRESS_BOOK_VCARD(ab)->address_list;
     while (list) {
 	address = LIBBALSA_ADDRESS(list->data);
 
@@ -330,33 +433,27 @@ libbalsa_address_book_vcard_store_address(LibBalsaAddressBook * ab,
 	return;
     }
 
-    fprintf(fp, "\nBEGIN:VCARD\n");
-    if (new_address->full_name) {
-	output = g_strdup_printf("FN:%s\n", new_address->full_name);
-	fprintf(fp, output);
-	g_free(output);
-    }
-    if (new_address->first_name && new_address->last_name) {
-	output =
-	    g_strdup_printf("N:%s;%s\n", new_address->last_name,
-			    new_address->first_name);
-	fprintf(fp, output);
-	g_free(output);
-    }
-    if (new_address->organization) {
-	output = g_strdup_printf("ORG:%s\n", new_address->organization);
-	fprintf(fp, output);
-	g_free(output);
-    }
-    list = new_address->address_list;
-    while (list) {
-	output =
-	    g_strdup_printf("EMAIL;INTERNET:%s\n", (gchar *) list->data);
-	fprintf(fp, output);
-	g_free(output);
-	list = g_list_next(list);
-    }
-    fprintf(fp, "END:VCARD\n");
+    fprintf(fp, "BEGIN:VCARD\n");
+    if (*new_address->full_name != '\0')
+	fprintf(fp, "FN:%s\n", new_address->full_name);
+    if (*new_address->last_name != '\0' && 
+	*new_address->first_name != '\0' && 
+	*new_address->middle_name != '\0')
+	fprintf(fp, "N:%s;%s;%s\n", new_address->last_name,
+		new_address->first_name, new_address->middle_name);
+    else if (*new_address->last_name != '\0' && 
+	     *new_address->first_name != '\0')
+	fprintf(fp, "N:%s;%s\n", new_address->last_name,
+		new_address->first_name);
+    else if (*new_address->first_name != '\0')
+	fprintf(fp, "N:;%s\n", new_address->first_name);
+    if (*new_address->organization != '\0')
+	fprintf(fp, "ORG:%s\n", new_address->organization);
+
+    for (list = new_address->address_list; list; list = g_list_next(list))
+	fprintf(fp, "EMAIL;INTERNET:%s\n", (gchar *) list->data);
+
+    fprintf(fp, "END:VCARD\n\n");
     fclose(fp);
 }
 
@@ -392,3 +489,125 @@ libbalsa_address_book_vcard_load_config(LibBalsaAddressBook * ab,
     if (LIBBALSA_ADDRESS_BOOK_CLASS(parent_class)->load_config)
 	LIBBALSA_ADDRESS_BOOK_CLASS(parent_class)->load_config(ab, prefix);
 }
+
+static GList *libbalsa_address_book_vcard_alias_complete(LibBalsaAddressBook * ab,
+							 const gchar * prefix, 
+							 gchar ** new_prefix)
+{
+    LibBalsaAddressBookVcard *vc;
+    GList *resa = NULL, *resb = NULL;
+    GList *res = NULL;
+    gchar *p1 = NULL, *p2 = NULL;
+    LibBalsaAddress *addr1, *addr2;
+
+    g_return_val_if_fail(LIBBALSA_IS_ADDRESS_BOOK_VCARD(ab), NULL);
+
+    vc = LIBBALSA_ADDRESS_BOOK_VCARD(ab);
+
+    if ( ab->expand_aliases == FALSE )
+	return NULL;
+
+    load_vcard_file(ab);
+
+    resa = g_completion_complete(vc->name_complete, (gchar*)prefix, &p1);
+    resb = g_completion_complete(vc->alias_complete, (gchar*)prefix, &p2);
+
+    if ( p1 && p2 ) {
+	if ( strlen(p1) > strlen(p2) ) {
+	    *new_prefix = p1;
+	    g_free(p2);
+	} else {
+	    *new_prefix = p2;
+	    g_free(p1);
+	}
+    } else {
+	*new_prefix = p1?p1:p2;
+    }
+
+    /*
+      Extract a list of addresses.
+      pick any of them if two addresses point to the same structure.
+      pick addr1 if it is available and there is no addr2 
+                    or it is smaller than addr1.
+      in other case, pick addr2 (one of addr1 or addr2 must be not-null).
+    */
+    while ( resa || resb ) {
+	addr1 = resa ? ((CompletionData*)resa->data)->address : NULL;
+	addr2 = resb ? ((CompletionData*)resb->data)->address : NULL;
+	
+	if (addr1 == addr2) {
+	    res = g_list_prepend(res, addr1);
+	    gtk_object_ref(GTK_OBJECT(addr1));
+	    resa = g_list_next(resa);
+	    resb = g_list_next(resb);
+	} else if (resa != NULL && 
+		   (resb == NULL || address_compare(addr1, addr2) > 0) ) {
+	    res = g_list_prepend(res, addr1);
+	    gtk_object_ref(GTK_OBJECT(addr1));
+	    resa = g_list_next(resa);
+	} else {
+	    res = g_list_prepend(res, addr2);
+	    gtk_object_ref(GTK_OBJECT(addr2));
+	    resb = g_list_next(resb);
+	}
+    }
+    res = g_list_reverse(res);
+
+    return res;
+}
+
+/*
+ * Create a new CompletionData
+ */
+static CompletionData *
+completion_data_new(LibBalsaAddress * address, gboolean alias)
+{
+    CompletionData *ret;
+
+    ret = g_new0(CompletionData, 1);
+
+    /*  gtk_object_ref(GTK_OBJECT(address)); */
+    ret->address = address;
+
+    if (alias)
+	ret->string = g_strdup(address->id);
+    else
+	ret->string = g_strdup(address->full_name);
+
+#ifdef CASE_INSENSITIVE_NAME
+    g_strup(ret->string);
+#endif
+
+    return ret;
+}
+
+/*
+ * Free a CompletionData
+ */
+static void
+completion_data_free(CompletionData * data)
+{
+    /*  gtk_object_unref(GTK_OBJECT(data->address)); */
+
+    g_free(data->string);
+    g_free(data);
+}
+
+/*
+ * The GCompletionFunc
+ */
+static gchar *
+completion_data_extract(CompletionData * data)
+{
+    return data->string;
+}
+
+static gint
+address_compare(LibBalsaAddress *a, LibBalsaAddress *b)
+{
+    g_return_val_if_fail(a != NULL, -1);
+    g_return_val_if_fail(b != NULL, 1);
+
+    return g_strcasecmp(a->full_name, b->full_name);
+}
+
