@@ -1,7 +1,6 @@
 /* -*-mode:c; c-style:k&r; c-basic-offset:4; -*- */
 /* Balsa E-Mail Client
- *
- * Copyright (C) 1997-2000 Stuart Parmenter and others,
+ * Copyright (C) 1997-2001 Stuart Parmenter and others,
  *                         See the file AUTHORS for a list.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -56,14 +55,18 @@ static void libbalsa_mailbox_real_load_config(LibBalsaMailbox * mailbox,
 static void message_status_changed_cb(LibBalsaMessage * message,
 				      LibBalsaMailbox * mb);
 
+static void libbalsa_mailbox_sync_backend_real(LibBalsaMailbox * mailbox);
 static LibBalsaMessage *translate_message(HEADER * cur);
 
 enum {
     OPEN_MAILBOX,
+    OPEN_MAILBOX_APPEND,
     CLOSE_MAILBOX,
     MESSAGE_STATUS_CHANGED,
     MESSAGE_NEW,
+    MESSAGES_NEW,
     MESSAGE_DELETE,
+    MESSAGES_DELETE,
     GET_MESSAGE_STREAM,
     CHECK,
     SET_UNREAD_MESSAGES_FLAG,
@@ -122,8 +125,15 @@ libbalsa_mailbox_class_init(LibBalsaMailboxClass * klass)
 		       object_class->type,
 		       GTK_SIGNAL_OFFSET(LibBalsaMailboxClass,
 					 open_mailbox),
-		       gtk_marshal_NONE__BOOL, GTK_TYPE_NONE, 1,
-		       GTK_TYPE_BOOL);
+		       gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
+
+    libbalsa_mailbox_signals[OPEN_MAILBOX_APPEND] =
+	gtk_signal_new("append-mailbox",
+		       GTK_RUN_LAST,
+		       object_class->type,
+		       GTK_SIGNAL_OFFSET(LibBalsaMailboxClass,
+					 open_mailbox_append),
+		       libbalsa_marshal_POINTER__NONE, GTK_TYPE_POINTER, 0);
 
     libbalsa_mailbox_signals[CLOSE_MAILBOX] =
 	gtk_signal_new("close-mailbox",
@@ -142,6 +152,15 @@ libbalsa_mailbox_class_init(LibBalsaMailboxClass * klass)
 		       gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1,
 		       LIBBALSA_TYPE_MESSAGE);
 
+    libbalsa_mailbox_signals[MESSAGES_NEW] =
+	gtk_signal_new("messages-new",
+		       GTK_RUN_FIRST,
+		       object_class->type,
+		       GTK_SIGNAL_OFFSET(LibBalsaMailboxClass,
+					 messages_new),
+		       gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1,
+		       GTK_TYPE_POINTER);
+
     libbalsa_mailbox_signals[MESSAGE_DELETE] =
 	gtk_signal_new("message-delete",
 		       GTK_RUN_FIRST,
@@ -151,6 +170,15 @@ libbalsa_mailbox_class_init(LibBalsaMailboxClass * klass)
 		       gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1,
 		       LIBBALSA_TYPE_MESSAGE);
 
+    libbalsa_mailbox_signals[MESSAGES_DELETE] =
+	gtk_signal_new("messages-delete",
+		       GTK_RUN_FIRST,
+		       object_class->type,
+		       GTK_SIGNAL_OFFSET(LibBalsaMailboxClass,
+					 messages_delete),
+		       gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1,
+		       GTK_TYPE_POINTER);
+
     libbalsa_mailbox_signals[SET_UNREAD_MESSAGES_FLAG] =
 	gtk_signal_new("set-unread-messages-flag",
 		       GTK_RUN_FIRST,
@@ -206,7 +234,9 @@ libbalsa_mailbox_class_init(LibBalsaMailboxClass * klass)
 
     klass->message_status_changed = NULL;
     klass->message_new = NULL;
+    klass->messages_new = NULL;
     klass->message_delete = NULL;
+    klass->messages_delete = NULL;
 
     klass->get_message_stream = NULL;
     klass->check = NULL;
@@ -222,6 +252,7 @@ libbalsa_mailbox_init(LibBalsaMailbox * mailbox)
 
     mailbox->config_prefix = NULL;
     mailbox->name = NULL;
+    mailbox->url = NULL;
     CLIENT_CONTEXT(mailbox) = NULL;
 
     mailbox->open_ref = 0;
@@ -233,6 +264,7 @@ libbalsa_mailbox_init(LibBalsaMailbox * mailbox)
     mailbox->message_list = NULL;
 
     mailbox->readonly = FALSE;
+    mailbox->mailing_list_address = NULL;
 }
 
 /* libbalsa_mailbox_destroy:
@@ -241,18 +273,25 @@ libbalsa_mailbox_init(LibBalsaMailbox * mailbox)
 static void
 libbalsa_mailbox_destroy(GtkObject * object)
 {
-    LibBalsaMailbox *mailbox = LIBBALSA_MAILBOX(object);
+    LibBalsaMailbox *mailbox;
+    g_return_if_fail(object);
 
-    if (!mailbox)
-	return;
+    mailbox = LIBBALSA_MAILBOX(object);
 
+    /* g_print("Destroying mailbox: %s\n", mailbox->name); */
     while (mailbox->open_ref > 0)
 	libbalsa_mailbox_close(mailbox);
 
-    g_free(mailbox->name);
-    mailbox->name = NULL;
+    if ( mailbox->mailing_list_address )
+	gtk_object_unref(GTK_OBJECT(mailbox->mailing_list_address));
+    mailbox->mailing_list_address = NULL;
+
     g_free(mailbox->config_prefix);
     mailbox->config_prefix = NULL;
+    g_free(mailbox->name);
+    mailbox->name = NULL;
+    g_free(mailbox->url);
+    mailbox->url = NULL;
 
     if (GTK_OBJECT_CLASS(parent_class)->destroy)
 	(*GTK_OBJECT_CLASS(parent_class)->destroy) (GTK_OBJECT(object));
@@ -270,9 +309,10 @@ libbalsa_mailbox_new_from_config(const gchar * prefix)
     gnome_config_push_prefix(prefix);
     type_str = gnome_config_get_string_with_default("Type", &got_default);
 
-    if (got_default == TRUE) {
+    if (got_default) {
 	libbalsa_information(LIBBALSA_INFORMATION_WARNING,
 			     _("Cannot load mailbox %s"), prefix);
+	gnome_config_pop_prefix();
 	return NULL;
     }
 
@@ -281,19 +321,25 @@ libbalsa_mailbox_new_from_config(const gchar * prefix)
 	libbalsa_information(LIBBALSA_INFORMATION_WARNING,
 			     _("No such mailbox type: %s"), type_str);
 	g_free(type_str);
+	gnome_config_pop_prefix();
 	return NULL;
     }
 
+    /* Handle Local mailboxes. 
+     * They are now separate classes for each type 
+     * FIXME: This should be removed in som efuture release.
+     */
+    if ( type == LIBBALSA_TYPE_MAILBOX_LOCAL ) {
+	gchar *path = gnome_config_get_string("Path");
+	type = libbalsa_mailbox_type_from_path(path);
+    }
     mailbox = gtk_type_new(type);
-    if (mailbox == NULL) {
+    if (mailbox == NULL)
 	libbalsa_information(LIBBALSA_INFORMATION_WARNING,
 			     _("Could not create a mailbox of type %s"),
 			     type_str);
-	g_free(type_str);
-	return NULL;
-    }
-
-    libbalsa_mailbox_load_config(mailbox, prefix);
+    else 
+	libbalsa_mailbox_load_config(mailbox, prefix);
 
     gnome_config_pop_prefix();
     g_free(type_str);
@@ -302,13 +348,13 @@ libbalsa_mailbox_new_from_config(const gchar * prefix)
 }
 
 void
-libbalsa_mailbox_open(LibBalsaMailbox * mailbox, gboolean append)
+libbalsa_mailbox_open(LibBalsaMailbox * mailbox)
 {
     g_return_if_fail(mailbox != NULL);
     g_return_if_fail(LIBBALSA_IS_MAILBOX(mailbox));
 
     gtk_signal_emit(GTK_OBJECT(mailbox),
-		    libbalsa_mailbox_signals[OPEN_MAILBOX], append);
+		    libbalsa_mailbox_signals[OPEN_MAILBOX]);
 }
 
 void
@@ -321,6 +367,29 @@ libbalsa_mailbox_close(LibBalsaMailbox * mailbox)
 		    libbalsa_mailbox_signals[CLOSE_MAILBOX]);
 }
 
+LibBalsaMailboxAppendHandle* 
+libbalsa_mailbox_open_append(LibBalsaMailbox * mailbox)
+{
+    LibBalsaMailboxAppendHandle* res = NULL;
+    g_return_val_if_fail(mailbox != NULL, NULL);
+    g_return_val_if_fail(LIBBALSA_IS_MAILBOX(mailbox), NULL);
+
+    gtk_signal_emit(GTK_OBJECT(mailbox),
+		    libbalsa_mailbox_signals[OPEN_MAILBOX_APPEND], &res);
+    return res;
+}
+
+int 
+libbalsa_mailbox_close_append(LibBalsaMailboxAppendHandle* handle)
+{
+    int ret;
+    g_return_val_if_fail(handle != NULL, -1);
+
+    ret = mx_close_mailbox(handle->context, NULL);
+    g_free(handle);
+    return ret;
+}
+
 void
 libbalsa_mailbox_set_unread_messages_flag(LibBalsaMailbox * mailbox,
 					  gboolean has_unread)
@@ -394,7 +463,8 @@ libbalsa_mailbox_get_message_stream(LibBalsaMailbox * mailbox,
 static void
 libbalsa_mailbox_real_close(LibBalsaMailbox * mailbox)
 {
-    int check;
+    static const int RETRIES_COUNT = 100;
+    int check, cnt;
 #ifdef DEBUG
     g_print("LibBalsaMailbox: Closing %s Refcount: %d\n", mailbox->name,
 	    mailbox->open_ref);
@@ -410,31 +480,34 @@ libbalsa_mailbox_real_close(LibBalsaMailbox * mailbox)
 
     if (mailbox->open_ref == 0) {
 	libbalsa_mailbox_free_messages(mailbox);
-	mailbox->messages = 0;
-	mailbox->total_messages = 0;
-	mailbox->unread_messages = 0;
 
 	/* now close the mail stream and expunge deleted
 	 * messages -- the expunge may not have to be done */
-	if (CLIENT_CONTEXT_OPEN(mailbox)) {
-	    /* We are careful to take/release locks in the correct order here */
-	    libbalsa_lock_mutt();
-	    while (
-		   (check =
-		    mx_close_mailbox(CLIENT_CONTEXT(mailbox), NULL))) {
-		libbalsa_unlock_mutt();
-		UNLOCK_MAILBOX(mailbox);
-		g_print
-		    ("libbalsa_mailbox_real_close: close failed, retrying...\n");
-		libbalsa_mailbox_check(mailbox);
-		LOCK_MAILBOX(mailbox);
-		libbalsa_lock_mutt();
-	    }
+	/* We are careful to take/release locks in the correct order here */
+	libbalsa_lock_mutt();
+	cnt = 0;
+	while( CLIENT_CONTEXT_OPEN(mailbox) && cnt < RETRIES_COUNT &&
+	       (check = mx_close_mailbox(CLIENT_CONTEXT(mailbox), NULL))) {
 	    libbalsa_unlock_mutt();
+	    UNLOCK_MAILBOX(mailbox);
+	    g_print
+		("libbalsa_mailbox_real_close: %d trial failed.\n", cnt);
+	    usleep(1000);
+	    libbalsa_mailbox_check(mailbox);
+	    LOCK_MAILBOX(mailbox);
+	    libbalsa_lock_mutt();
+	    cnt++;
+	}
+	libbalsa_unlock_mutt();
+	if(cnt>=RETRIES_COUNT)
+	    g_print("libbalsa_mailbox_real_close: changes to %s lost.\n",
+		    mailbox->name);
+	
+	if(CLIENT_CONTEXT(mailbox)) {
 	    free(CLIENT_CONTEXT(mailbox));
 	    CLIENT_CONTEXT(mailbox) = NULL;
 	}
-    }
+    } else libbalsa_mailbox_sync_backend_real(mailbox);
 
     UNLOCK_MAILBOX(mailbox);
 }
@@ -458,12 +531,21 @@ static void
 libbalsa_mailbox_real_save_config(LibBalsaMailbox * mailbox,
 				  const gchar * prefix)
 {
+    gchar *tmp;
     g_return_if_fail(LIBBALSA_IS_MAILBOX(mailbox));
 
     gnome_config_set_string("Type",
 			    gtk_type_name(GTK_OBJECT_TYPE(mailbox)));
     gnome_config_set_string("Name", mailbox->name);
 
+    if ( mailbox->mailing_list_address ) {
+	tmp = libbalsa_address_to_gchar(mailbox->mailing_list_address, 0);
+	gnome_config_set_string("MailingListAddress", tmp);
+	g_free(tmp);
+    } else {
+	gnome_config_clean_key("MailingListAddress");
+    }
+
     g_free(mailbox->config_prefix);
     mailbox->config_prefix = g_strdup(prefix);
 
@@ -473,6 +555,8 @@ static void
 libbalsa_mailbox_real_load_config(LibBalsaMailbox * mailbox,
 				  const gchar * prefix)
 {
+    gboolean def;
+    gchar *address;
     g_return_if_fail(LIBBALSA_IS_MAILBOX(mailbox));
 
     g_free(mailbox->config_prefix);
@@ -480,12 +564,26 @@ libbalsa_mailbox_real_load_config(LibBalsaMailbox * mailbox,
 
     g_free(mailbox->name);
     mailbox->name = gnome_config_get_string("Name=Mailbox");
+
+    if ( mailbox->mailing_list_address ) 
+	gtk_object_unref(GTK_OBJECT(mailbox->mailing_list_address));
+    address = gnome_config_get_string_with_default("MailingListAddress", &def);
+    if ( def == TRUE ) {
+	mailbox->mailing_list_address = NULL;
+    } else {
+	mailbox->mailing_list_address = libbalsa_address_new_from_string(address);
+    }
+    g_free(address);
 }
 
 /*
  * private 
  * PS: called by mail_progress_notify_cb:
  * loads incrementally new messages, if any.
+ *  Mailbox lock MUST NOT BE HELD before calling this function.
+ *  gdk_lock MUST BE HELD before calling this function because it is called
+ *  from both threading and not threading code and we want to be on the safe
+ *  side.
  */
 void
 libbalsa_mailbox_load_messages(LibBalsaMailbox * mailbox)
@@ -493,10 +591,12 @@ libbalsa_mailbox_load_messages(LibBalsaMailbox * mailbox)
     glong msgno;
     LibBalsaMessage *message;
     HEADER *cur = 0;
+    GList *messages=NULL;
 
     if (CLIENT_CONTEXT_CLOSED(mailbox))
 	return;
 
+    LOCK_MAILBOX(mailbox);
     for (msgno = mailbox->messages; mailbox->new_messages > 0; msgno++) {
 	cur = CLIENT_CONTEXT(mailbox)->hdrs[msgno];
 
@@ -549,6 +649,11 @@ libbalsa_mailbox_load_messages(LibBalsaMailbox * mailbox)
 	if (cur->replied)
 	    message->flags |= LIBBALSA_MESSAGE_FLAG_REPLIED;
 
+        /* set the length */
+#ifdef MESSAGE_COPY_CONTENT
+        message->length = cur->content->length;
+        message->lines_len = cur->lines;
+#endif
 	/* take over the ownership */
 	mailbox->message_list =
 	    g_list_append(mailbox->message_list, message);
@@ -557,20 +662,25 @@ libbalsa_mailbox_load_messages(LibBalsaMailbox * mailbox)
 
 	mailbox->new_messages--;
 
-	gtk_signal_emit(GTK_OBJECT(mailbox),
-			libbalsa_mailbox_signals[MESSAGE_NEW], message);
+	messages=g_list_append(messages, message);
+    }
+    UNLOCK_MAILBOX(mailbox);
+
+    if(messages!=NULL){
+      gtk_signal_emit(GTK_OBJECT(mailbox),
+			libbalsa_mailbox_signals[MESSAGES_NEW], messages);
+      g_list_free(messages);
     }
 
-    if (mailbox->unread_messages > 0)
-	libbalsa_mailbox_set_unread_messages_flag(mailbox, TRUE);
-    else
-	libbalsa_mailbox_set_unread_messages_flag(mailbox, FALSE);
+    libbalsa_mailbox_set_unread_messages_flag(mailbox,
+					      mailbox->unread_messages > 0);
 }
 
 /* libbalsa_mailbox_free_messages:
    removes all the messages from the mailbox.
    Messages are unref'ed and not directly destroyed because they migt
    be ref'ed from somewhere else.
+   Mailbox lock MUST BE HELD before calling this function.
 */
 void
 libbalsa_mailbox_free_messages(LibBalsaMailbox * mailbox)
@@ -580,47 +690,56 @@ libbalsa_mailbox_free_messages(LibBalsaMailbox * mailbox)
 
     list = g_list_first(mailbox->message_list);
 
+    if(list){
+      gtk_signal_emit(GTK_OBJECT(mailbox),
+		      libbalsa_mailbox_signals[MESSAGES_DELETE], list);
+    }
+
     while (list) {
 	message = list->data;
 	list = list->next;
 
-	gtk_signal_emit(GTK_OBJECT(mailbox),
-			libbalsa_mailbox_signals[MESSAGE_DELETE], message);
 	message->mailbox = NULL;
 	gtk_object_unref(GTK_OBJECT(message));
     }
 
     g_list_free(mailbox->message_list);
     mailbox->message_list = NULL;
+    mailbox->messages = 0;
+    mailbox->total_messages = 0;
+    mailbox->unread_messages = 0;
 }
 
-LibBalsaMailboxType libbalsa_mailbox_valid(gchar * filename)
+GtkType libbalsa_mailbox_type_from_path(const gchar * filename)
 {
     struct stat st;
-    LibBalsaMailboxType ret;
+    GtkType ret = 0;
 
+    /*
+     * FIXME: Needed? if the file doesn't exist won't
+     * we get some unknown value from mx_get_magic 
+     */
     if (stat(filename, &st) == -1)
-	return MAILBOX_UNKNOWN;
+	return 0;
 
     libbalsa_lock_mutt();
     switch (mx_get_magic(filename)) {
     case M_MBOX:
-	ret = MAILBOX_MBOX;
+	ret = LIBBALSA_TYPE_MAILBOX_MBOX;
 	break;
     case M_MMDF:
-	ret = MAILBOX_MBOX;
+	ret = LIBBALSA_TYPE_MAILBOX_MBOX;
 	break;
     case M_MH:
-	ret = MAILBOX_MH;
+	ret = LIBBALSA_TYPE_MAILBOX_MH;
 	break;
     case M_MAILDIR:
-	ret = MAILBOX_MAILDIR;
+	ret = LIBBALSA_TYPE_MAILBOX_MAILDIR;
 	break;
     case M_IMAP:
-	ret = MAILBOX_IMAP;
+	ret = LIBBALSA_TYPE_MAILBOX_IMAP;
 	break;
-    default:
-	ret = MAILBOX_UNKNOWN;
+    case M_KENDRA: /* We don't support KENDRA */
 	break;
     }
     libbalsa_unlock_mutt();
@@ -628,55 +747,82 @@ LibBalsaMailboxType libbalsa_mailbox_valid(gchar * filename)
     return ret;
 }
 
-/* libbalsa_mailbox_commit_changes:
-   commits the changes to the file. note that the msg numbers are changed 
-   after commit so one has to re-read messages from mutt structures.
-   Actually, re-reading is a wrong approach because it slows down balsa
-   like hell. I know, I tried it (re-reading).  */
-gint
-libbalsa_mailbox_commit_changes(LibBalsaMailbox * mailbox)
+static void
+libbalsa_mailbox_sync_backend_real(LibBalsaMailbox * mailbox)
 {
     GList *message_list;
     GList *tmp_message_list;
     LibBalsaMessage *current_message;
-    gint res = 0;
+    GList *p=NULL;
+
+    message_list = mailbox->message_list;
+    while (message_list) {
+	current_message = LIBBALSA_MESSAGE(message_list->data);
+	tmp_message_list = message_list->next;
+	if (current_message->flags & LIBBALSA_MESSAGE_FLAG_DELETED) {
+	    p=g_list_append(p, current_message);
+	}
+	message_list = tmp_message_list;
+    }
+    if(p){
+	gtk_signal_emit(GTK_OBJECT(mailbox),
+			libbalsa_mailbox_signals[MESSAGES_DELETE],
+			p);
+	g_list_free(p);
+    }
 
-    libbalsa_mailbox_open(mailbox, FALSE);
-
-    /* examine all the message in the mailbox */
     message_list = mailbox->message_list;
     while (message_list) {
 	current_message = LIBBALSA_MESSAGE(message_list->data);
 	tmp_message_list = message_list->next;
 	if (current_message->flags & LIBBALSA_MESSAGE_FLAG_DELETED) {
-	    gtk_signal_emit(GTK_OBJECT(mailbox),
-			    libbalsa_mailbox_signals[MESSAGE_DELETE],
-			    current_message);
 	    current_message->mailbox = NULL;
 	    gtk_object_unref(GTK_OBJECT(current_message));
 	    mailbox->message_list =
 		g_list_remove_link(mailbox->message_list, message_list);
 	}
 	message_list = tmp_message_list;
-
+	
     }
+}
+
+/* libbalsa_mailbox_sync_backend:
+   synchronizes the backend with LibBalsaMailbox.
+
+   NOTE: this is not a proper mailbox commit function.  When
+   implementing proper one note that the msg numbers are changed after
+   commit so one has to re-read messages from mutt structures.
+   Actually, re-reading is a wrong approach because it slows down
+   balsa like hell. I know, I tried it (re-reading).
+*/
+gint
+libbalsa_mailbox_sync_backend(LibBalsaMailbox * mailbox)
+{
+    gint res = 0;
+
+    /* only open mailboxes can be commited; lock it instead of opening */
+    g_return_val_if_fail(CLIENT_CONTEXT_OPEN(mailbox), 0);
+    LOCK_MAILBOX_RETURN_VAL(mailbox, 0);
+    libbalsa_mailbox_sync_backend_real(mailbox);
+
 #if 0
     libbalsa_lock_mutt();
-    res = 0;			/* FIXME: mx_sync_mailbox (CLIENT_CONTEXT(mailbox), NULL); */
+    res = 0; /* FIXME: mx_sync_mailbox (CLIENT_CONTEXT(mailbox), NULL); */
     if (res) {
 	libbalsa_mailbox_free_messages(mailbox);
-	mailbox->messages = 0;
-	mailbox->total_messages = 0;
-	mailbox->unread_messages = 0;
 	libbalsa_mailbox_load_messages(mailbox);
     }
     libbalsa_unlock_mutt();
 #endif
-    libbalsa_mailbox_close(mailbox);
+    UNLOCK_MAILBOX(mailbox);
     return res;
 }
 
-/* internal c-client translation */
+
+/* internal c-client translation:
+ * mutt lists can cantain null adresses for address strings like
+ * "To: Dear Friends,". We do remove them.
+ */
 static LibBalsaMessage *
 translate_message(HEADER * cur)
 {
@@ -698,20 +844,22 @@ translate_message(HEADER * cur)
     message->from = libbalsa_address_new_from_libmutt(cenv->from);
     message->sender = libbalsa_address_new_from_libmutt(cenv->sender);
     message->reply_to = libbalsa_address_new_from_libmutt(cenv->reply_to);
+    message->dispnotify_to = 
+	libbalsa_address_new_from_libmutt(cenv->dispnotify_to);
 
     for (addy = cenv->to; addy; addy = addy->next) {
 	addr = libbalsa_address_new_from_libmutt(addy);
-	message->to_list = g_list_append(message->to_list, addr);
+	if(addr) message->to_list = g_list_append(message->to_list, addr);
     }
 
     for (addy = cenv->cc; addy; addy = addy->next) {
 	addr = libbalsa_address_new_from_libmutt(addy);
-	message->cc_list = g_list_append(message->cc_list, addr);
+	if(addr) message->cc_list = g_list_append(message->cc_list, addr);
     }
 
     for (addy = cenv->bcc; addy; addy = addy->next) {
 	addr = libbalsa_address_new_from_libmutt(addy);
-	message->bcc_list = g_list_append(message->bcc_list, addr);
+	if(addr) message->bcc_list = g_list_append(message->bcc_list, addr);
     }
 
     /* Get fcc from message */
@@ -730,11 +878,22 @@ translate_message(HEADER * cur)
 	    SKIPWS(p);
 
 	    message->in_reply_to = g_strdup(p);
+	} else if (g_strncasecmp ("In-Reply-To:", tmp->data, 12) == 0){
+	    p = tmp->data + 12;
+	    while(*p!='\0'&& *p!='<')p++;
+	    if(*p!='\0'){
+		message->in_reply_to = g_strdup (p);
+		p=message->in_reply_to;
+		while(*p!='\0' && *p!='>')p++;
+		if(*p=='>')*(p+1)='\0';
+	    }
 	}
 	tmp = tmp->next;
     }
 
-    message->subject = g_strdup(cenv->subject);
+#ifdef MESSAGE_COPY_CONTENT
+    message->subj = g_strdup(cenv->subject);
+#endif
     message->message_id = g_strdup(cenv->message_id);
 
     for (tmp = cenv->references; tmp != NULL; tmp = tmp->next) {
@@ -744,6 +903,33 @@ translate_message(HEADER * cur)
 
     /* more! */
 
+    if(cenv->references!=NULL){
+	LIST* p=cenv->references;
+	while(p!=NULL){
+	    message->references_for_threading = 
+		g_list_prepend(message->references_for_threading, 
+			       g_strdup(p->data));
+            p=p->next;
+	}
+    }
+
+#if 0
+    /* According to  RFC 1036 (section 2.2.5), MessageIDs in References header
+     * must be in the oldest first order; the direct parent should be last. 
+     * It seems, however, some MUAs ignore this rule.
+     * For example, one of them adds the direct parent to the head of
+     * the References header.
+     */
+    if(message->in_reply_to != NULL &&
+       message->references_for_threading != NULL &&
+       1<g_list_length(message->references_for_threading) &&
+       strcmp(message->in_reply_to, 
+	      (g_list_first(message->references_for_threading))->data)==0){
+	GList *foo=message->references_for_threading;
+        message->references_for_threading=g_list_remove(foo, foo->data);
+    }
+#endif
+
     return message;
 }
 
