@@ -1,7 +1,6 @@
 /* -*-mode:c; c-style:k&r; c-basic-offset:4; -*- */
 /* Balsa E-Mail Client
- *
- * Copyright (C) 1997-2000 Stuart Parmenter and others,
+ * Copyright (C) 1997-2001 Stuart Parmenter and others,
  *                         See the file AUTHORS for a list.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -20,6 +19,17 @@
  * 02111-1307, USA.
  */
 
+/* DESIGN NOTES.
+   MESSAGE_COPY_CONTENT define is an attempt to reduce memory usage of balsa.
+   When it is defined, The message date is stored in one place only (in
+   libmutt structures). This should reduce memory usage to some extent.
+   However, it is not implemented very extensively at the present moment
+   and the memory usage reduction is hardly noticeable.
+   - Lack of inline functions in C increases program complexity. This cost
+   can be accepted.
+   - thorough analysis of memory usage is needed.
+*/
+   
 #include "config.h"
 
 #include <glib.h>
@@ -106,7 +116,8 @@ libbalsa_message_init(LibBalsaMessage * message)
     message->from = NULL;
     message->sender = NULL;
     message->reply_to = NULL;
-    message->subject = NULL;
+    message->dispnotify_to = NULL;
+    message->subj = NULL;
     message->to_list = NULL;
     message->cc_list = NULL;
     message->bcc_list = NULL;
@@ -116,6 +127,7 @@ libbalsa_message_init(LibBalsaMessage * message)
     message->message_id = NULL;
     message->body_ref = 0;
     message->body_list = NULL;
+    message->references_for_threading = NULL;
 }
 
 
@@ -223,6 +235,10 @@ libbalsa_message_destroy(GtkObject * object)
 	gtk_object_unref(GTK_OBJECT(message->reply_to));
 	message->reply_to = NULL;
     }
+    if(message->dispnotify_to) {
+	gtk_object_unref(GTK_OBJECT(message->dispnotify_to));
+	message->dispnotify_to = NULL;
+    }
 
     g_list_foreach(message->to_list, (GFunc) gtk_object_unref, NULL);
     g_list_free(message->to_list);
@@ -236,9 +252,10 @@ libbalsa_message_destroy(GtkObject * object)
     g_list_free(message->bcc_list);
     message->bcc_list = NULL;
 
-    g_free(message->subject);
-    message->subject = NULL;
-
+#if MESSAGE_COPY_CONTENT
+    g_free(message->subj);
+    message->subj = NULL;
+#endif
     for (list = message->references; list; list = list->next) {
 	if (list->data)
 	    g_free(list->data);
@@ -255,6 +272,19 @@ libbalsa_message_destroy(GtkObject * object)
 
     libbalsa_message_body_free(message->body_list);
     message->body_list = NULL;
+
+    if(message->references_for_threading!=NULL) {
+	GList *list=message->references_for_threading;
+	for(; list; list=g_list_next(list)){
+	    if(list->data)
+		g_free(list->data);
+	}
+	g_list_free (message->references_for_threading);
+	message->references_for_threading=NULL;
+    }
+
+    if (GTK_OBJECT_CLASS(parent_class)->destroy)
+	(*GTK_OBJECT_CLASS(parent_class)->destroy) (GTK_OBJECT(object));
 }
 
 const gchar *
@@ -264,24 +294,47 @@ libbalsa_message_pathname(LibBalsaMessage * message)
     return CLIENT_CONTEXT(message->mailbox)->hdrs[message->msgno]->path;
 }
 
-const gchar *
-libbalsa_message_charset(LibBalsaMessage * message)
+static const gchar *
+libbalsa_message_body_charset(LibBalsaMessageBody * body)
 {
     gchar *charset = NULL;
 
-    LibBalsaMessageBody *body = message->body_list;
     while (body) {
-	charset = libbalsa_message_body_get_parameter(body, "charset");
+	libbalsa_lock_mutt();
+	charset = mutt_get_parameter("charset", body->mutt_body->parameter);
+	libbalsa_unlock_mutt();
+
+	if (charset)
+	    break;
+	
+	if (body->parts)
+	    charset = (gchar *)libbalsa_message_body_charset(body->parts);
 
 	if (charset)
 	    break;
 
 	body = body->next;
     }
-
     return charset;
 }
 
+/* Note: libbalsa_message_charset returns a pointer to the charset field or
+ * NULL, but does NOT make a copy of an existing string! */
+const gchar *
+libbalsa_message_charset(LibBalsaMessage * message)
+{
+    LibBalsaMessageBody *body;
+
+    g_return_val_if_fail(message != NULL, NULL);
+    body = message->body_list;
+    g_return_val_if_fail(body != NULL, NULL);
+
+    if (body->charset)
+	return body->charset;
+    else
+	return libbalsa_message_body_charset(body);
+}
+
 /* message_user_hdrs:
    returns allocated GList containing (header=>value) ALL headers pairs
    as generated by g_strsplit.
@@ -309,6 +362,10 @@ libbalsa_message_user_hdrs(LibBalsaMessage * message)
     gchar **pair;
 
     g_return_val_if_fail(message->mailbox, NULL);
+    if(CLIENT_CONTEXT(message->mailbox)->hdrs == NULL) 
+	/* oops, mutt closed the mailbox on error, we should do the same */
+	return NULL;
+
     cur = CLIENT_CONTEXT(message->mailbox)->hdrs[message->msgno];
     g_assert(cur != NULL);
     g_assert(cur->env != NULL);
@@ -356,85 +413,112 @@ libbalsa_message_user_hdrs(LibBalsaMessage * message)
     return res;
 }
 
+
+gboolean
+libbalsa_message_move(LibBalsaMessage * message, LibBalsaMailbox * dest)
+{
+    if (message->mailbox->readonly) {
+	libbalsa_information(
+	    LIBBALSA_INFORMATION_ERROR,
+	    _("Source mailbox (%s) is readonly. Cannot move message"),
+	    message->mailbox->name);
+	return FALSE;
+    }
+
+    if (libbalsa_message_copy (message, dest)) {
+        libbalsa_message_delete (message);
+        return TRUE;
+    } else
+        return FALSE;
+}
+
+gboolean
+libbalsa_messages_move (GList* messages, LibBalsaMailbox* dest)
+{
+    g_return_val_if_fail(messages, FALSE);
+
+    if (LIBBALSA_MESSAGE(messages->data)->mailbox->readonly) {
+	libbalsa_information(
+	    LIBBALSA_INFORMATION_ERROR,
+	    _("Source mailbox (%s) is readonly. Cannot move messages"),
+	    LIBBALSA_MESSAGE(messages->data)->mailbox->name);
+	return FALSE;
+    }
+
+    if (libbalsa_messages_copy (messages, dest)) {
+        libbalsa_messages_delete (messages);
+        return TRUE;
+    } else {
+        return FALSE;
+    }
+}
+
 /* FIXME: look at the flags for mutt_append_message */
 gboolean
 libbalsa_message_copy(LibBalsaMessage * message, LibBalsaMailbox * dest)
 {
     HEADER *cur;
-
+    LibBalsaMailboxAppendHandle* handle;
     g_return_val_if_fail(message->mailbox, FALSE);
     RETURN_VAL_IF_CONTEXT_CLOSED(message->mailbox, FALSE);
 
     cur = CLIENT_CONTEXT(message->mailbox)->hdrs[message->msgno];
 
-    libbalsa_mailbox_open(dest, FALSE);
+    handle = libbalsa_mailbox_open_append(dest);
 
-    if (!CLIENT_CONTEXT(dest) || dest->readonly == TRUE) {
-	libbalsa_information(LIBBALSA_INFORMATION_WARNING,
-			     _
-			     ("Couldn't open destination mailbox (%s) for copying"),
-			     dest->name);
+    if (!handle) {
+	libbalsa_information(
+	    LIBBALSA_INFORMATION_WARNING,
+	    _("Couldn't open destination mailbox (%s) for copying"),
+	    dest->name);
 	return FALSE;
     }
 
     libbalsa_lock_mutt();
-    mutt_append_message(CLIENT_CONTEXT(dest),
+    mutt_append_message(handle->context,
 			CLIENT_CONTEXT(message->mailbox), cur, 0, 0);
     libbalsa_unlock_mutt();
 
-    dest->total_messages++;
-    if (message->flags & LIBBALSA_MESSAGE_FLAG_NEW)
-	dest->unread_messages++;
-
-    libbalsa_mailbox_close(dest);
+    libbalsa_mailbox_close_append(handle);
+    libbalsa_mailbox_check(dest);
     return TRUE;
 }
 
+/* libbalsa_messages_copy:
+   makes an assumption that all the messages come from the same mailbox.
+*/
 gboolean
-libbalsa_message_move(LibBalsaMessage * message, LibBalsaMailbox * dest)
+libbalsa_messages_copy (GList * messages, LibBalsaMailbox * dest)
 {
     HEADER *cur;
+    LibBalsaMessage *message;
+    GList *p;
+    LibBalsaMailboxAppendHandle* handle;
 
-    g_return_val_if_fail(message != NULL, FALSE);
+    g_return_val_if_fail(messages != NULL, FALSE);
     g_return_val_if_fail(dest != NULL, FALSE);
-    g_return_val_if_fail(message->mailbox,FALSE);
-    RETURN_VAL_IF_CONTEXT_CLOSED(message->mailbox, FALSE);
 
-    cur = CLIENT_CONTEXT(message->mailbox)->hdrs[message->msgno];
-
-    libbalsa_mailbox_open(dest, TRUE);
-
-    if (!CLIENT_CONTEXT(dest) || dest->readonly == TRUE) {
-	libbalsa_information(LIBBALSA_INFORMATION_WARNING,
-			     _
-			     ("Couldn't open destination mailbox (%s) for writing"),
-			     dest->name);
-	return FALSE;
-    }
-    if (message->mailbox->readonly == TRUE) {
-	libbalsa_information(LIBBALSA_INFORMATION_WARNING,
-			     _
-			     ("Source mailbox (%s) is readonly. Cannot move message"),
-			     message->mailbox->name);
+    handle = libbalsa_mailbox_open_append(dest);
+    if (!handle) {
+	libbalsa_information(
+	    LIBBALSA_INFORMATION_ERROR,
+	    _("Couldn't open destination mailbox (%s) for writing"),
+	    dest->name);
 	return FALSE;
     }
 
     libbalsa_lock_mutt();
-
-    mutt_parse_mime_message(CLIENT_CONTEXT(message->mailbox), cur);
-
-    mutt_append_message(CLIENT_CONTEXT(dest),
-			CLIENT_CONTEXT(message->mailbox), cur, 0, 0);
-
+    for(p=messages; p; 	p=g_list_next(p)) {
+	message=(LibBalsaMessage *)(p->data);
+	cur = CLIENT_CONTEXT(message->mailbox)->hdrs[message->msgno];
+	mutt_parse_mime_message(CLIENT_CONTEXT(message->mailbox), cur);
+	mutt_append_message(handle->context,
+			    CLIENT_CONTEXT(message->mailbox), cur, 0, 0);
+    }
     libbalsa_unlock_mutt();
 
-    dest->total_messages++;
-
-    if (message->flags & LIBBALSA_MESSAGE_FLAG_NEW)
-	dest->unread_messages++;
-
-    libbalsa_mailbox_close(dest);
-    libbalsa_message_delete(message);
+    libbalsa_mailbox_close_append(handle);
+    libbalsa_mailbox_check(dest);
     return TRUE;
 }
 
@@ -495,7 +579,8 @@ libbalsa_message_real_set_read_flag(LibBalsaMessage * message,
 
     } else if (!set) {
 	libbalsa_lock_mutt();
-	mutt_set_flag(CLIENT_CONTEXT(message->mailbox), cur, M_READ, TRUE);
+	mutt_set_flag(CLIENT_CONTEXT(message->mailbox), cur, M_READ, FALSE);
+	mutt_set_flag(CLIENT_CONTEXT(message->mailbox), cur, M_OLD, FALSE);
 	libbalsa_unlock_mutt();
 
 	message->flags |= LIBBALSA_MESSAGE_FLAG_NEW;
@@ -652,6 +737,19 @@ libbalsa_message_delete(LibBalsaMessage * message)
 		    libbalsa_message_signals[SET_DELETED], TRUE);
 }
 
+void
+libbalsa_messages_delete(GList * messages)
+{
+    LibBalsaMessage * message;
+    g_return_if_fail(messages != NULL);
+
+    while(messages){
+      message=(LibBalsaMessage *)(messages->data);
+      gtk_signal_emit(GTK_OBJECT(message),
+  	  	      libbalsa_message_signals[SET_DELETED], TRUE);
+      messages=g_list_next(messages);
+    }
+}
 
 void
 libbalsa_message_undelete(LibBalsaMessage * message)
@@ -688,24 +786,29 @@ mime_content_type2str(int contenttype)
 }
 #endif
 
-void
+/* libbalsa_message_body_ref:
+   references the body of given message.
+   NO OP for 'loose' messages (i.e not associated with any mailbox).
+   returns TRUE for success, FALSE for failure (broken IMAP connection etc).
+*/
+gboolean
 libbalsa_message_body_ref(LibBalsaMessage * message)
 {
     LibBalsaMessageBody *body;
     HEADER *cur;
     MESSAGE *msg;
 
-    if (!message)
-	return;
+    g_return_val_if_fail(message, FALSE);
+    if (!message->mailbox) return FALSE;
 
     if (message->body_ref > 0) {
 	message->body_ref++;
-	return;
+	return TRUE;
     }
-
+    g_return_val_if_fail(CLIENT_CONTEXT(message->mailbox)->hdrs, FALSE);
     cur = CLIENT_CONTEXT(message->mailbox)->hdrs[message->msgno];
     if (cur == NULL)
-	return;
+	return FALSE;
 
     /*
      * load message body
@@ -716,7 +819,7 @@ libbalsa_message_body_ref(LibBalsaMessage * message)
 
     if (!msg) {
 	message->body_ref--;
-	return;
+	return FALSE;
     }
 
     fseek(msg->fp, cur->content->offset, 0);
@@ -771,6 +874,7 @@ libbalsa_message_body_ref(LibBalsaMessage * message)
      */
     if (message->flags & LIBBALSA_MESSAGE_FLAG_NEW)
 	libbalsa_message_read(message);
+    return TRUE;
 }
 
 
@@ -794,6 +898,10 @@ libbalsa_message_has_attachment(LibBalsaMessage * message)
     gboolean ret;
     HEADER *msg_header;
 
+    g_return_val_if_fail(LIBBALSA_IS_MESSAGE(message), FALSE);
+    g_return_val_if_fail(message->mailbox, FALSE);
+    g_return_val_if_fail(CLIENT_CONTEXT(message->mailbox)->hdrs, FALSE);
+
     msg_header = CLIENT_CONTEXT(message->mailbox)->hdrs[message->msgno];
 
     /* FIXME: Can be simplified into 1 if */
@@ -826,6 +934,36 @@ libbalsa_message_date_to_gchar(LibBalsaMessage * message,
     return g_strdup(rettime);
 }
 
+gchar *
+libbalsa_message_size_to_gchar (LibBalsaMessage * message, gboolean lines)
+{
+    gchar retsize[32];
+    glong length;   /* byte len */
+    gint lines_len; /* line len */
+
+    g_return_val_if_fail(message != NULL, NULL);
+    length    = LIBBALSA_MESSAGE_GET_LENGTH(message);
+    lines_len = LIBBALSA_MESSAGE_GET_LINES(message);
+    /* lines is int, length is long */
+    if (lines)
+        g_snprintf (retsize, sizeof(retsize), "%d", lines_len);
+    else {
+        if (length <= 32768) {
+            g_snprintf (retsize, sizeof(retsize), "%ld", length);
+        } else if (length <= (100*1024)) {
+            float tmp = (float)length/1024.0;
+            g_snprintf (retsize, sizeof(retsize), "%.1fK", tmp);
+        } else if (length <= (1024*1024)) {
+            g_snprintf (retsize, sizeof(retsize), "%ldK", length/1024);
+        } else {
+            float tmp = (float)length/(1024.0*1024.0);
+            g_snprintf (retsize, sizeof(retsize), "%.1fM", tmp);
+        }
+    }
+
+    return g_strdup(retsize);
+}
+
 static gchar *
 ADDRESS_to_gchar(const ADDRESS * addr)
 {
@@ -857,40 +995,46 @@ libbalsa_message_append_part(LibBalsaMessage * message,
     }
 }
 
-/* libbalsa_message_get_text_content:
-   returns the message content as single string. 
-   When modifying make sure it works properly for both messages in mailboxes
-   as well as messages created on the fly.
-
-   FIXME: This is a simple-minded version of this function, just to get 
-   the stable release out. The full blown code should mention attachments.
-   Or we could create separate function for it, so printing would be 
-   even nicer.
-   content2reply() is also kind of specialized function that should be
-   implemented in a more general fashion.  
+/* libbalsa_message_set_dispnotify:
+   sets a disposition notify to a given address
+   address can be NULL.
 */
-gchar *
-libbalsa_message_get_text_content(LibBalsaMessage * msg, gint line_len)
+void
+libbalsa_message_set_dispnotify(LibBalsaMessage *message, 
+				LibBalsaAddress *address)
 {
-    gchar *res;
-    GString *str;
-    g_return_val_if_fail(msg, NULL);
-
-    if (msg->mailbox) {
-	libbalsa_message_body_ref(msg);
-	str = content2reply(msg, NULL, line_len);
-	libbalsa_message_body_unref(msg);
-    } else {
-	LibBalsaMessageBody *body = msg->body_list;
-	str = g_string_new("");
-	while (body) {
-	    if (body->buffer)
-		str = g_string_append(str, body->buffer);
-	    body = body->next;
-	}
-	libbalsa_wrap_string(str->str, line_len);
-    }
-    res = str->str;
-    g_string_free(str, FALSE);
-    return res;
+    g_return_if_fail(message);
+    if(message->dispnotify_to) 
+	gtk_object_unref(GTK_OBJECT(message->dispnotify_to));
+    message->dispnotify_to = address;
+    if(address)
+	gtk_object_ref(GTK_OBJECT(message->dispnotify_to));
 }
+
+#ifndef MESSAGE_COPY_CONTENT
+/* libbalsa_message_get_subject:
+   get constant pointer to the subject of the message; 
+*/
+const gchar*
+libbalsa_message_get_subject(LibBalsaMessage* msg)
+{
+    if(msg->mailbox) {
+	/* g_print("Returning libmutt's pointer\n"); */
+	return CLIENT_CONTEXT(msg->mailbox)->hdrs[msg->msgno]->env->subject;
+    } else
+	return msg->subj;
+}
+
+
+guint
+libbalsa_message_get_lines(LibBalsaMessage* msg)
+{
+    return CLIENT_CONTEXT(msg->mailbox)->hdrs[msg->msgno]->lines;
+}
+glong
+libbalsa_message_get_length(LibBalsaMessage* msg)
+{
+    return CLIENT_CONTEXT(msg->mailbox)->hdrs[msg->msgno]->content->length;
+}
+
+#endif
