@@ -1,7 +1,6 @@
 /* -*-mode:c; c-style:k&r; c-basic-offset:4; -*- */
 /* Balsa E-Mail Client
- *
- * Copyright (C) 1997-2000 Stuart Parmenter and others,
+ * Copyright (C) 1997-2001 Stuart Parmenter and others,
  *                         See the file AUTHORS for a list.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -30,6 +29,11 @@
 
 #include "libbalsa.h"
 
+#if ENABLE_ESMTP
+#include <auth-client.h>
+#endif
+
+/* #define MESSAGE_COPY_CONTENT 1 */
 #define LIBBALSA_TYPE_MESSAGE                      (libbalsa_message_get_type())
 #define LIBBALSA_MESSAGE(obj)                      (GTK_CHECK_CAST(obj, LIBBALSA_TYPE_MESSAGE, LibBalsaMessage))
 #define LIBBALSA_MESSAGE_CLASS(klass)              (GTK_CHECK_CLASS_CAST(klass, LIBBALSA_TYPE_MESSAGE, LibBalsaMessageClass))
@@ -65,14 +69,24 @@ struct _LibBalsaMessage {
     /* message composition date */
     time_t date;
 
-    /* from, sender, and reply addresses */
+    /* from, sender, reply, and disposition notification addresses */
     LibBalsaAddress *from;
     LibBalsaAddress *sender;
     LibBalsaAddress *reply_to;
+    LibBalsaAddress *dispnotify_to;
 
-    /* subject line */
-    gchar *subject;
-
+    /* subject line; we still need it here for sending;
+     * although _SET_SUBJECT might resolve it(?) 
+     * but we can set to to NULL unless there is no mailbox, like
+     * on sending. */
+    gchar *subj;
+#ifdef MESSAGE_COPY_CONTENT
+#define LIBBALSA_MESSAGE_GET_SUBJECT(m) ((m)->subj)
+#else
+#define LIBBALSA_MESSAGE_GET_SUBJECT(m) libbalsa_message_get_subject(m)
+#endif
+#define LIBBALSA_MESSAGE_SET_SUBJECT(m,s) \
+        { g_free((m)->subj); (m)->subj = (s); }
     /* primary, secondary, and blind recipent lists */
     GList *to_list;
     GList *cc_list;
@@ -83,6 +97,7 @@ struct _LibBalsaMessage {
 
     /* replied message ID's */
     GList *references;
+    GList *references_for_threading; /* oldest first */
 
     /* replied message ID; from address on date */
     gchar *in_reply_to;
@@ -94,6 +109,16 @@ struct _LibBalsaMessage {
     guint body_ref;
     LibBalsaMessageBody *body_list;
     /*  GList *body_list; */
+
+#if MESSAGE_COPY_CONTENT
+    glong length;   /* byte len */
+    gint lines_len; /* line len */
+#define LIBBALSA_MESSAGE_GET_LENGTH(m)  ((m)->length)
+#define LIBBALSA_MESSAGE_GET_LINES(m) ((m)->lines_len)
+#else
+#define LIBBALSA_MESSAGE_GET_LENGTH(m) libbalsa_message_get_length(m)
+#define LIBBALSA_MESSAGE_GET_LINES(m)  libbalsa_message_get_lines(m)
+#endif
 };
 
 struct _LibBalsaMessageClass {
@@ -124,11 +149,16 @@ gboolean libbalsa_message_copy(LibBalsaMessage * message,
 			       LibBalsaMailbox * dest);
 gboolean libbalsa_message_move(LibBalsaMessage * message,
 			       LibBalsaMailbox * mailbox);
+gboolean libbalsa_messages_move(GList * messages,
+				LibBalsaMailbox * dest);
+gboolean libbalsa_messages_copy(GList * messages,
+                                LibBalsaMailbox * dest);
 void libbalsa_message_clear_flags(LibBalsaMessage * message);
 
 void libbalsa_message_read(LibBalsaMessage * message);
 void libbalsa_message_unread(LibBalsaMessage * message);
 void libbalsa_message_delete(LibBalsaMessage * message);
+void libbalsa_messages_delete(GList * messages);
 void libbalsa_message_undelete(LibBalsaMessage * message);
 
 void libbalsa_message_flag(LibBalsaMessage * message);
@@ -140,7 +170,7 @@ void libbalsa_message_reply(LibBalsaMessage * message);
 void libbalsa_message_append_part(LibBalsaMessage * message,
 				  LibBalsaMessageBody * body);
 
-void libbalsa_message_body_ref(LibBalsaMessage * message);
+gboolean libbalsa_message_body_ref(LibBalsaMessage * message);
 void libbalsa_message_body_unref(LibBalsaMessage * message);
 
 gboolean balsa_send_message(LibBalsaMessage * message,
@@ -149,8 +179,22 @@ gboolean balsa_postpone_message(LibBalsaMessage * message,
 				LibBalsaMailbox * draftbox,
 				LibBalsaMessage * reply_message,
 				gchar * fcc);
-gboolean libbalsa_message_send(LibBalsaMessage * message,
-			       LibBalsaMailbox * outbox, gint encoding);
+void libbalsa_message_queue(LibBalsaMessage* message, 
+			    LibBalsaMailbox* outbox, LibBalsaMailbox* fccbox,
+			    gint encoding);
+#if ENABLE_ESMTP
+gboolean libbalsa_message_send(LibBalsaMessage* message,
+			       LibBalsaMailbox* outbox,  
+			       LibBalsaMailbox* fccbox,
+			       gint encoding, gchar* smtp_server,
+			       auth_context_t smtp_authctx,
+			       gint tls_mode);
+#else
+gboolean libbalsa_message_send(LibBalsaMessage* message,
+			       LibBalsaMailbox* outbox,  
+			       LibBalsaMailbox* fccbox,
+			       gint encoding);
+#endif
 gboolean libbalsa_message_postpone(LibBalsaMessage * message,
 				   LibBalsaMailbox * draftbox,
 				   LibBalsaMessage * reply_message,
@@ -161,12 +205,24 @@ gboolean libbalsa_message_postpone(LibBalsaMessage * message,
  */
 gchar *libbalsa_message_date_to_gchar(LibBalsaMessage * message,
 				      const gchar * date_string);
+gchar *libbalsa_message_size_to_gchar(LibBalsaMessage * message,
+                                      gboolean lines);
+
 const gchar *libbalsa_message_pathname(LibBalsaMessage * message);
 const gchar *libbalsa_message_charset(LibBalsaMessage * message);
 gboolean libbalsa_message_has_attachment(LibBalsaMessage * message);
 
 GList *libbalsa_message_user_hdrs(LibBalsaMessage * message);
-gchar *libbalsa_message_get_text_content(LibBalsaMessage * msg,
-					 gint line_len);
 
+void libbalsa_message_set_dispnotify(LibBalsaMessage *message, 
+				     LibBalsaAddress *address);
+
+/* use LIBBALSA_MESSAGE_GET_SUBJECT() macro, we may optimize this
+   function out if we find a way.
+*/
+#ifndef MESSAGE_COPY_CONTENT
+const gchar* libbalsa_message_get_subject(LibBalsaMessage* message);
+guint libbalsa_message_get_lines(LibBalsaMessage* msg);
+glong libbalsa_message_get_length(LibBalsaMessage* msg);
+#endif
 #endif				/* __LIBBALSA_MESSAGE_H__ */
