@@ -16,6 +16,8 @@
  *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */ 
 
+#define _SENDLIB_C 1
+
 #include "mutt.h"
 #include "mutt_curses.h"
 #include "rfc2047.h"
@@ -568,7 +570,6 @@ static CONTENT *mutt_get_content_info (const char *fname)
 {
   CONTENT *info;
   FILE *fp;
-  CHARSET_MAP *cm;
   int ch, from=0, whitespace=0, dot=0, linelen=0;
 
   if ((fp = fopen (fname, "r")) == NULL)
@@ -578,12 +579,6 @@ static CONTENT *mutt_get_content_info (const char *fname)
     return (NULL);
   }
 
-  {
-    CHARSET *cs;
-
-    cm = (cs = mutt_get_charset(Charset)) ? cs->map : 0;
-  }
-
   info = safe_calloc (1, sizeof (CONTENT));
   while ((ch = fgetc (fp)) != EOF)
   {
@@ -656,7 +651,7 @@ static CONTENT *mutt_get_content_info (const char *fname)
       if (ch == ' ') whitespace++;
       info->ascii++;
     }
-    if (cm && mutt_unicode_char (cm, ch) & -128)
+    if (ch>127)
       info->nonasc = 1;
     if (linelen > 1) dot = 0;
     if (ch != ' ' && ch != '\t') whitespace = 0;
@@ -762,17 +757,15 @@ static int lookup_mime_type (char *d, const char *s)
 
 static char *set_text_charset (CONTENT *info)
 {
-  CHARSET *cs;
-
   /* if charset is unknown assume low bytes are ascii compatible */
-
+#if 0
   if ((Charset == NULL || mutt_strcasecmp (Charset, "us-ascii") == 0)
       && info->hibin)
     return ("unknown-8bit");
 
   if (((cs = mutt_get_charset (Charset)) && cs->map) ? info->nonasc : info->hibin)
     return (Charset);
-
+#endif
   return ("us-ascii");
 }
 
@@ -1332,6 +1325,15 @@ int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, int mode)
     mutt_write_address_list (env->mail_followup_to, fp, 18);
   }
 
+  /* BALSA: Handle dispnotify_to field */
+  if (env->dispnotify_to)
+  {
+    buffer[0] = 0;
+    rfc822_write_address (buffer, sizeof (buffer), env->dispnotify_to);
+    fprintf (fp, "Disposition-Notification-To: %s\n", buffer);
+  }
+  /* BALSA: end of dispnotify_to field handling code */
+
   if (mode <= 0)
   {
     if (env->references)
@@ -1364,27 +1366,37 @@ int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, int mode)
   return (ferror (fp) == 0 ? 0 : -1);
 }
 
+
+/* BALSA: new encode header to match new rfc2047 */
 static void encode_headers (LIST *h)
 {
-  char tmp[LONG_STRING];
+  char *tmp;
   char *p;
-  size_t len;
-
+  int i;
+  
   for (; h; h = h->next)
   {
-    if ((p = strchr (h->data, ':')))
-    {
-      *p++ = 0;
-      SKIPWS (p);
-      snprintf (tmp, sizeof (tmp), "%s: ", h->data);
-      len = mutt_strlen (tmp);
-      rfc2047_encode_string (tmp + len, sizeof (tmp) - len, (unsigned char *) p);
-      safe_free ((void **) &h->data);
-      h->data = safe_strdup (tmp);
-    }
+    if (!(p = strchr (h->data, ':')))
+      continue;
+
+    i = p - h->data;
+    ++p; SKIPWS (p);
+    tmp = safe_strdup (p);
+
+    if (!tmp)
+      continue;
+    
+    rfc2047_encode_string (&tmp);
+    safe_realloc ((void **) &h->data, 
+                  mutt_strlen (h->data) + 2 + mutt_strlen (tmp) + 1);
+
+    sprintf (h->data + i, ": %s", NONULL (tmp));  /* __SPRINTF_CHECKED__ */
+    
+    safe_free ((void **) &tmp);
   }
 }
 
+
 const char *mutt_fqdn(short may_hide_host)
 {
   char *p = NULL, *q;
@@ -1862,21 +1874,21 @@ void mutt_prepare_envelope (ENVELOPE *env)
 #endif
 
   /* Take care of 8-bit => 7-bit conversion. */
-  rfc2047_encode_adrlist (env->to);
-  rfc2047_encode_adrlist (env->cc);
-  rfc2047_encode_adrlist (env->from);
-  rfc2047_encode_adrlist (env->mail_followup_to);
-  rfc2047_encode_adrlist (env->reply_to);
+  rfc2047_encode_adrlist (env->to, NULL);
+  rfc2047_encode_adrlist (env->cc, NULL);
+  rfc2047_encode_adrlist (env->from, NULL);
+  rfc2047_encode_adrlist (env->mail_followup_to, NULL);
+  rfc2047_encode_adrlist (env->reply_to, NULL);
+  /* BALSA: Handle dispnotify_to field */
+  rfc2047_encode_adrlist (env->dispnotify_to, NULL);
 
+  /* BALSA: updated to new rfc2047 */
   if (env->subject)
-  {
-    rfc2047_encode_string (buffer, sizeof (buffer) - 1,
-			   (unsigned char *) env->subject);
-    safe_free ((void **) &env->subject);
-    env->subject = safe_strdup (buffer);
-  }
+    {
+      rfc2047_encode_string (&env->subject);
+    }
   encode_headers (env->userhdrs);
-
+  
   if (!env->message_id)
     env->message_id = mutt_gen_msgid ();
 }
