@@ -1,6 +1,6 @@
 /* -*-mode:c; c-style:k&r; c-basic-offset:4; -*- */
 /* Balsa E-Mail Client
- * Copyright (C) 1997-2000 Stuart Parmenter and others,
+ * Copyright (C) 1997-2001 Stuart Parmenter and others,
  *                         See the file AUTHORS for a list.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -24,39 +24,49 @@
 
 #include "balsa-icons.h"
 
+#include "pixmaps/all_headers.xpm"
+#include "pixmaps/close.xpm"
+#include "pixmaps/flag_new.xpm"
+#include "pixmaps/flagged.xpm"
+
+#include "pixmaps/identity.xpm"
+#include "pixmaps/identity_menu.xpm"
+
+#include "pixmaps/mark_all.xpm"
+
+#include "pixmaps/next_unread.xpm"
+#include "pixmaps/next_unread_menu.xpm"
+
+#include "pixmaps/reply_to_all.xpm"
+#include "pixmaps/reply_to_all_menu.xpm"
+
+#include "pixmaps/draftbox.xpm"
 #include "pixmaps/inbox.xpm"
 #include "pixmaps/outbox.xpm"
+#include "pixmaps/sentbox.xpm"
 #include "pixmaps/trash.xpm"
 
-#include "pixmaps/dir_closed.xpm"
-#include "pixmaps/dir_open.xpm"
-
 #include "pixmaps/tray_empty.xpm"
 #include "pixmaps/tray_full.xpm"
 
+#include "pixmaps/dir_closed.xpm"
+#include "pixmaps/dir_open.xpm"
+
 #include "pixmaps/replied.xpm"
 #include "pixmaps/forwarded.xpm"
-
 #include "pixmaps/envelope.xpm"
-
-#include "pixmaps/arrow.xpm"
 #include "pixmaps/multipart.xpm"
 
-#include "pixmaps/flagged.xpm"
-
-#include "pixmaps/reply_to_all.xpm"
-#include "pixmaps/reply_to_all_menu.xpm"
-#include "pixmaps/next_unread.xpm"
-#include "pixmaps/next_unread_menu.xpm"
-
 typedef struct _BalsaIcon BalsaIcon;
 struct _BalsaIcon {
     GdkPixmap *p;
     GdkBitmap *b;
 };
 
+static BalsaIcon draftbox;
 static BalsaIcon inbox;
 static BalsaIcon outbox;
+static BalsaIcon sentbox;
 static BalsaIcon trash;
 
 static BalsaIcon tray_empty;
@@ -67,10 +77,7 @@ static BalsaIcon dir_closed;
 
 static BalsaIcon replied;
 static BalsaIcon forwarded;
-
 static BalsaIcon envelope;
-
-static BalsaIcon arrow;
 static BalsaIcon multipart;
 
 static void
@@ -83,8 +90,10 @@ create_icon(gchar ** data, GdkPixmap ** pmap, GdkBitmap ** bmap)
 void
 balsa_icons_init(void)
 {
+    create_icon(draftbox_xpm, &draftbox.p, &draftbox.b);
     create_icon(inbox_xpm, &inbox.p, &inbox.b);
     create_icon(outbox_xpm, &outbox.p, &outbox.b);
+    create_icon(sentbox_xpm, &sentbox.p, &sentbox.b);
     create_icon(trash_xpm, &trash.p, &trash.b);
 
     create_icon(tray_empty_xpm, &tray_empty.p, &tray_empty.b);
@@ -95,10 +104,7 @@ balsa_icons_init(void)
 
     create_icon(replied_xpm, &replied.p, &replied.b);
     create_icon(forwarded_xpm, &forwarded.p, &forwarded.b);
-
     create_icon(envelope_xpm, &envelope.p, &envelope.b);
-
-    create_icon(arrow_xpm, &arrow.p, &arrow.b);
     create_icon(multipart_xpm, &multipart.p, &multipart.b);
 }
 
@@ -106,36 +112,24 @@ GdkPixmap *
 balsa_icon_get_pixmap(BalsaIconName name)
 {
     switch (name) {
-    case BALSA_ICON_INBOX:
-	return inbox.p;
-    case BALSA_ICON_OUTBOX:
-	return outbox.p;
-    case BALSA_ICON_TRASH:
-	return trash.p;
+	case BALSA_ICON_DRAFTBOX:	return draftbox.p;
+	case BALSA_ICON_INBOX:		return inbox.p;
+	case BALSA_ICON_OUTBOX:		return outbox.p;
+	case BALSA_ICON_SENTBOX:	return sentbox.p;
+	case BALSA_ICON_TRASH:		return trash.p;
 
-    case BALSA_ICON_TRAY_EMPTY:
-	return tray_empty.p;
-    case BALSA_ICON_TRAY_FULL:
-	return tray_full.p;
+	case BALSA_ICON_TRAY_EMPTY:	return tray_empty.p;
+	case BALSA_ICON_TRAY_FULL:	return tray_full.p;
 
-    case BALSA_ICON_DIR_CLOSED:
-	return dir_closed.p;
-    case BALSA_ICON_DIR_OPEN:
-	return dir_open.p;
+	case BALSA_ICON_DIR_CLOSED:	return dir_closed.p;
+	case BALSA_ICON_DIR_OPEN:	return dir_open.p;
 
-    case BALSA_ICON_REPLIED:
-	return replied.p;
-    case BALSA_ICON_FORWARDED:
-	return forwarded.p;
-
-    case BALSA_ICON_ENVELOPE:
-	return envelope.p;
-
-    case BALSA_ICON_ARROW:
-	return arrow.p;
-    case BALSA_ICON_MULTIPART:
-	return multipart.p;
+	case BALSA_ICON_REPLIED:	return replied.p;
+	case BALSA_ICON_FORWARDED:	return forwarded.p;
+	case BALSA_ICON_ENVELOPE:	return envelope.p;
+	case BALSA_ICON_MULTIPART:	return multipart.p;
     }
+
     return NULL;
 }
 
@@ -143,37 +137,24 @@ GdkBitmap *
 balsa_icon_get_bitmap(BalsaIconName name)
 {
     switch (name) {
-    case BALSA_ICON_INBOX:
-	return inbox.b;
-    case BALSA_ICON_OUTBOX:
-	return outbox.b;
-    case BALSA_ICON_TRASH:
-	return trash.b;
+	case BALSA_ICON_DRAFTBOX:	return draftbox.b;
+	case BALSA_ICON_INBOX:		return inbox.b;
+	case BALSA_ICON_OUTBOX:		return outbox.b;
+	case BALSA_ICON_SENTBOX:	return sentbox.b;
+	case BALSA_ICON_TRASH:		return trash.b;
 
-    case BALSA_ICON_TRAY_EMPTY:
-	return tray_empty.b;
-    case BALSA_ICON_TRAY_FULL:
-	return tray_full.b;
+	case BALSA_ICON_TRAY_EMPTY:	return tray_empty.b;
+	case BALSA_ICON_TRAY_FULL:	return tray_full.b;
 
-    case BALSA_ICON_DIR_CLOSED:
-	return dir_closed.b;
-    case BALSA_ICON_DIR_OPEN:
-	return dir_open.b;
-
-    case BALSA_ICON_REPLIED:
-	return replied.b;
-    case BALSA_ICON_FORWARDED:
-	return forwarded.b;
-
-    case BALSA_ICON_ENVELOPE:
-	return envelope.b;
-
-    case BALSA_ICON_ARROW:
-	return arrow.b;
-    case BALSA_ICON_MULTIPART:
-	return multipart.b;
+	case BALSA_ICON_DIR_CLOSED:	return dir_closed.b;
+	case BALSA_ICON_DIR_OPEN:	return dir_open.b;
 
+	case BALSA_ICON_REPLIED:	return replied.b;
+	case BALSA_ICON_FORWARDED:	return forwarded.b;
+	case BALSA_ICON_ENVELOPE:	return envelope.b;
+	case BALSA_ICON_MULTIPART:	return multipart.b;
     }
+
     return NULL;
 }
 
@@ -194,13 +175,28 @@ register_balsa_pixmap(gchar * name, gchar ** data, guint xsize, guint ysize)
 void
 register_balsa_pixmaps(void)
 {
-    register_balsa_pixmap(BALSA_PIXMAP_MAIL_RPL_ALL, reply_to_all_xpm, 24, 24);
-    register_balsa_pixmap(BALSA_PIXMAP_MAIL_RPL_ALL_MENU,
-			  reply_to_all_menu_xpm, 16, 15);
-
-    register_balsa_pixmap (BALSA_PIXMAP_NEXT_UNREAD, next_unread_xpm, 24, 24);
-    register_balsa_pixmap (BALSA_PIXMAP_NEXT_UNREAD_MENU, 
-                           next_unread_menu_xpm, 16, 15);
-
-    register_balsa_pixmap (BALSA_PIXMAP_FLAGGED, flagged_xpm, 16, 16);
+    register_balsa_pixmap(BALSA_PIXMAP_SHOW_ALL_HEADERS, all_headers_xpm,
+			  24, 24);
+    register_balsa_pixmap(BALSA_PIXMAP_SMALL_CLOSE, close_xpm,
+			  9, 9);
+    register_balsa_pixmap(BALSA_PIXMAP_ENVELOPE, envelope_xpm,
+			  16, 16);
+    register_balsa_pixmap(BALSA_PIXMAP_FLAG_UNREAD, flag_new_xpm,
+			  24, 24);
+    register_balsa_pixmap(BALSA_PIXMAP_FLAGGED, flagged_xpm,
+			  16, 16);
+    register_balsa_pixmap(BALSA_PIXMAP_IDENTITY, identity_xpm,
+			  24, 24);
+    register_balsa_pixmap(BALSA_PIXMAP_IDENTITY_MENU, identity_menu_xpm,
+			  16, 16);
+    register_balsa_pixmap(BALSA_PIXMAP_MARK_ALL_MSGS, mark_all_xpm,
+			  24, 24);
+    register_balsa_pixmap(BALSA_PIXMAP_NEXT_UNREAD, next_unread_xpm,
+			  24, 24);
+    register_balsa_pixmap(BALSA_PIXMAP_NEXT_UNREAD_MENU, next_unread_menu_xpm,
+			  16, 16);
+    register_balsa_pixmap(BALSA_PIXMAP_MAIL_RPL_ALL, reply_to_all_xpm,
+			  24, 24);
+    register_balsa_pixmap(BALSA_PIXMAP_MAIL_RPL_ALL_MENU, reply_to_all_menu_xpm,
+			  16, 16);
 }
