@@ -22,6 +22,9 @@
 #ifndef __BALSA_MBLIST_H__
 #define __BALSA_MBLIST_H__
 
+#include <gtk/gtk.h>
+#include "mailbox-node.h"
+
 #define BALSA_TYPE_MBLIST          (balsa_mblist_get_type ())
 #define BALSA_MBLIST(obj)          GTK_CHECK_CAST (obj, BALSA_TYPE_MBLIST, BalsaMBList)
 #define BALSA_MBLIST_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, BALSA_TYPE_MBLIST, BalsaMBListClass)
@@ -33,32 +36,25 @@ typedef struct _BalsaMBListClass BalsaMBListClass;
 struct _BalsaMBList {
     GtkCTree ctree;
 
-    /* list of mailbox watched */
-    GList *watched_mailbox;
+    /* ugly, but there's no way to know this with the ctree itself ... */
+    GtkCTreeNode *currently_selected_ctree_node;
+    
     /* store the style of unread mailboxes */
     GtkStyle *unread_mailbox_style;
     /* shall the number of messages be displayed ? */
-    gboolean display_content_info;
-
-#ifdef BALSA_USE_THREADS
-    /* list of mailboxes for the updating thread */
-    GList *update_list;
-#endif
+    gboolean display_info;
 };
 
 struct _BalsaMBListClass {
     GtkCTreeClass parent_class;
-
-    void (*select_mailbox) (BalsaMBList * bmblist,
-			    LibBalsaMailbox * mailbox, gint row,
-			    GdkEventButton *);
 };
 
 GtkType balsa_mblist_get_type(void);
 
 GtkWidget *balsa_mblist_new(void);
 
-void balsa_mblist_redraw(BalsaMBList * bmbl);
+void balsa_mblist_repopulate(BalsaMBList * bmbl);
+void mblist_default_signal_bindings(BalsaMBList * tree);
 
 void balsa_mblist_have_new(BalsaMBList * bmbl);
 void balsa_mblist_update_mailbox(BalsaMBList * mblist,
@@ -66,4 +62,14 @@ void balsa_mblist_update_mailbox(BalsaMBList * mblist,
 gboolean balsa_mblist_focus_mailbox(BalsaMBList * bmbl,
 				    LibBalsaMailbox * mailbox);
 
+GList *mblist_find_all_unread_mboxes(void);
+LibBalsaMailbox* mblist_find_mbox_by_name(BalsaMBList * mblist,
+					     const char* name);
+void mblist_open_mailbox(LibBalsaMailbox * mailbox);
+void mblist_close_mailbox(LibBalsaMailbox * mailbox);
+BalsaMailboxNode* mblist_get_selected_node(BalsaMBList *mblist);
+BalsaMailboxNode* mblist_get_node_by_mailbox(BalsaMBList *mblist,
+					     LibBalsaMailbox * mailbox);
+gboolean mblist_remove_mailbox_node(BalsaMBList *mblist,
+				    BalsaMailboxNode* mbnode);
 #endif
