@@ -1,6 +1,6 @@
 /* -*-mode:c; c-style:k&r; c-basic-offset:4; -*- */
 /* Balsa E-Mail Client
- * Copyright (C) 1997-2000 Stuart Parmenter and others,
+ * Copyright (C) 1997-2001 Stuart Parmenter and others,
  *                         See the file AUTHORS for a list.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -33,21 +33,26 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <signal.h>
 
 #ifdef BALSA_USE_THREADS
 #include <pthread.h>
 #endif
 
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
+
 #include "balsa-app.h"
 #include "balsa-icons.h"
-#include "mblist-window.h"
-#include "balsa-mblist.h"
+#include "balsa-index.h"
 #include "main-window.h"
 #include "libbalsa.h"
-#include "misc.h"
+#include "mailbox-node.h"
 #include "save-restore.h"
-#include "main.h"
+#include "sendmsg-window.h"
 #include "information.h"
+#include "pop3.h"
 
 #include "libinit_balsa/init_balsa.h"
 
@@ -59,7 +64,6 @@ pthread_t get_mail_thread;
 pthread_t send_mail;
 pthread_mutex_t mailbox_lock;
 pthread_mutex_t send_messages_lock;
-pthread_mutex_t appbar_lock;
 int checking_mail;
 int sending_mail;
 int mail_thread_pipes[2];
@@ -77,13 +81,14 @@ int updating_mblist;
 /* Semaphore to prevent dual use of appbar progressbar */
 int updating_progressbar;
 
-static void threads_init(gboolean init);
+static void threads_init(void);
+static void threads_destroy(void);
 #endif				/* BALSA_USE_THREADS */
 
 static void balsa_init(int argc, char **argv);
 static void config_init(void);
 static void mailboxes_init(void);
-static void empty_trash(void);
+static void balsa_cleanup(void);
 static gint balsa_kill_session(GnomeClient * client, gpointer client_data);
 static gint balsa_save_session(GnomeClient * client, gint phase,
 			       GnomeSaveStyle save_style, gint is_shutdown,
@@ -108,11 +113,15 @@ balsa_init(int argc, char **argv)
 	 N_("Get new mail on startup"), NULL},
 	{"compose", 'm', POPT_ARG_STRING, &(balsa_app.compose_email),
 	 0, N_("Compose a new email to EMAIL@ADDRESS"), "EMAIL@ADDRESS"},
+	{"attach", 'a', POPT_ARG_STRING, &(balsa_app.attach_file),
+	 0, N_("Attach file at PATH"), "PATH"},
 	{"open-mailbox", 'o', POPT_ARG_STRING, &(cmd_line_open_mailboxes),
 	 0, N_("Opens MAILBOXNAME"), N_("MAILBOXNAME")},
 	{"open-unread-mailbox", 'u', POPT_ARG_NONE,
 	 &(cmd_open_unread_mailbox), 0,
 	 N_("Opens first unread mailbox"), NULL},
+	{"debug-pop", 'd', POPT_ARG_NONE, &PopDebug, 0, 
+	 N_("Debug POP3 connection"), NULL},
 	{NULL, '\0', 0, NULL, 0}	/* end the list */
     };
 
@@ -163,11 +172,9 @@ check_special_mailboxes(void)
 static void
 config_init(void)
 {
-    config_load();
-    if (check_special_mailboxes()) {
-	g_warning("*** Could not load basic mailboxes!\n");
+    while(!config_load() || check_special_mailboxes()) {
+	g_warning("*** Could not load basic configuration!\n");
 	balsa_init_begin();
-	/*return; */
     }
 }
 
@@ -183,43 +190,44 @@ mailboxes_init(void)
 
 
 #ifdef BALSA_USE_THREADS
-void
-threads_init(gboolean init)
+static void
+threads_init(void)
 {
-    if (init) {
-	g_thread_init(NULL);
-	pthread_mutex_init(&mailbox_lock, NULL);
-	pthread_mutex_init(&send_messages_lock, NULL);
-	pthread_mutex_init(&appbar_lock, NULL);
-	checking_mail = 0;
-	updating_mblist = 0;
-	sending_mail = 0;
-	updating_progressbar = 0;
-	if (pipe(mail_thread_pipes) < 0) {
-	    g_log("BALSA Init", G_LOG_LEVEL_DEBUG,
-		  "Error opening pipes.\n");
-	}
-	mail_thread_msg_send = g_io_channel_unix_new(mail_thread_pipes[1]);
-	mail_thread_msg_receive =
-	    g_io_channel_unix_new(mail_thread_pipes[0]);
-	g_io_add_watch(mail_thread_msg_receive, G_IO_IN,
-		       (GIOFunc) mail_progress_notify_cb, NULL);
-
-	if (pipe(send_thread_pipes) < 0) {
-	    g_log("BALSA Init", G_LOG_LEVEL_DEBUG,
-		  "Error opening pipes.\n");
-	}
-	send_thread_msg_send = g_io_channel_unix_new(send_thread_pipes[1]);
-	send_thread_msg_receive =
-	    g_io_channel_unix_new(send_thread_pipes[0]);
-	g_io_add_watch(send_thread_msg_receive, G_IO_IN,
-		       (GIOFunc) send_progress_notify_cb, NULL);
-    } else {
-	pthread_mutex_destroy(&mailbox_lock);
-	pthread_mutex_destroy(&send_messages_lock);
-	pthread_mutex_destroy(&appbar_lock);
+    g_thread_init(NULL);
+    pthread_mutex_init(&mailbox_lock, NULL);
+    pthread_mutex_init(&send_messages_lock, NULL);
+    checking_mail = 0;
+    updating_mblist = 0;
+    sending_mail = 0;
+    updating_progressbar = 0;
+    if (pipe(mail_thread_pipes) < 0) {
+	g_log("BALSA Init", G_LOG_LEVEL_DEBUG,
+	      "Error opening pipes.\n");
     }
+    mail_thread_msg_send = g_io_channel_unix_new(mail_thread_pipes[1]);
+    mail_thread_msg_receive =
+	g_io_channel_unix_new(mail_thread_pipes[0]);
+    g_io_add_watch(mail_thread_msg_receive, G_IO_IN,
+		   (GIOFunc) mail_progress_notify_cb, NULL);
+    
+    if (pipe(send_thread_pipes) < 0) {
+	g_log("BALSA Init", G_LOG_LEVEL_DEBUG,
+	      "Error opening pipes.\n");
+    }
+    send_thread_msg_send = g_io_channel_unix_new(send_thread_pipes[1]);
+    send_thread_msg_receive =
+	g_io_channel_unix_new(send_thread_pipes[0]);
+    g_io_add_watch(send_thread_msg_receive, G_IO_IN,
+		   (GIOFunc) send_progress_notify_cb, NULL);
 }
+
+static void
+threads_destroy(void)
+{
+    pthread_mutex_destroy(&mailbox_lock);
+    pthread_mutex_destroy(&send_messages_lock);
+}
+
 #endif				/* BALSA_USE_THREADS */
 
 /* initial_open_mailboxes:
@@ -243,64 +251,6 @@ initial_open_unread_mailboxes()
     return FALSE;
 }
 
-/* decode_and_strdup:
-   decodes given URL string up to the delimiter and places the
-   eos pointer in newstr if supplied (eos==NULL if end of string was reached)
-*/
-typedef void (*field_setter)(BalsaSendmsg *d, const gchar*, const gchar*);
-static gchar* decode_and_strdup(const gchar*str, int delim, gchar** newstr)
-{
-    gchar num[3];
-    GString *s = g_string_new(NULL);
-    /* eos points to the character after the last to parse */
-    gchar *eos = strchr(str, delim); 
-
-    if(!eos) eos = (gchar*)str + strlen(str);
-    while(str<eos) {
-	switch(*str) {
-	case '+':
-	    s = g_string_append_c(s, ' ');
-	    str++;
-	    break;
-	case '%':
-	    if(str+2<eos) {
-		strncpy(num, str+1, 2); num[2] = 0;
-		s = g_string_append_c(s, strtol(num,NULL,16));
-	    }
-	    str+=3;
-	    break;
-	default:
-	    s = g_string_append_c(s, *str++);
-	}
-    }
-    if(newstr) *newstr = *eos ? eos+1 : NULL;
-    eos = s->str;
-    g_string_free(s,FALSE);
-    return eos;
-}
-    
-/* process_url:
-   extracts all characters until NUL or question mark; parse later fields
-   of format 'key'='value' with ampersands as separators.
-*/ 
-static void process_url(const char *url, field_setter func, void *data)
-{
-    gchar * ptr, *to, *key, *val;
-
-    to = decode_and_strdup(url,'?', &ptr);
-    func(data, "to", to);
-    g_free(to);
-    while(ptr) {
-	key = decode_and_strdup(ptr,'=', &ptr);
-	if(ptr) {
-	    val = decode_and_strdup(ptr,'&', &ptr);
-	    func(data, key, val);
-	    g_free(val);
-	}
-	g_free(key);
-    }
-}
-
 /* -------------------------- main --------------------------------- */
 int
 main(int argc, char *argv[])
@@ -309,23 +259,26 @@ main(int argc, char *argv[])
     GnomeClient *client;
     gchar *default_icon;
 #ifdef GTKHTML_HAVE_GCONF
-    GConfError *gconf_error;
+    GError *gconf_error;
 #endif
 
+#ifdef ENABLE_NLS
     /* Initialize the i18n stuff */
     bindtextdomain(PACKAGE, GNOMELOCALEDIR);
     textdomain(PACKAGE);
+    setlocale(LC_CTYPE, gnome_i18n_get_language());
+#endif
 
 #ifdef BALSA_USE_THREADS
     /* initiate thread mutexs, variables */
-    threads_init(TRUE);
+    threads_init();
 #endif
 
     balsa_init(argc, argv);
 
 #ifdef GTKHTML_HAVE_GCONF
     if (!gconf_init(argc, argv, &gconf_error))
-	gconf_error_destroy(gconf_error);
+	g_error_free(gconf_error);
     gconf_error = NULL;
 #endif
 
@@ -366,6 +319,7 @@ main(int argc, char *argv[])
 
     window = balsa_window_new();
     balsa_app.main_window = BALSA_WINDOW(window);
+    gtk_signal_connect(GTK_OBJECT(window), "destroy", balsa_cleanup, NULL);
 
     /* session management */
     client = gnome_master_client();
@@ -376,13 +330,28 @@ main(int argc, char *argv[])
 
     gdk_rgb_init();
 
-    if (balsa_app.compose_email) {
+    if (balsa_app.compose_email || balsa_app.attach_file) {
 	BalsaSendmsg *snd;
 	snd = sendmsg_window_new(window, NULL, SEND_NORMAL);
-	if(strncmp(balsa_app.compose_email, "mailto:", 7) == 0)
-	    process_url(balsa_app.compose_email+7, 
-			sendmsg_window_set_field, snd);
-	else sendmsg_window_set_field(snd,"to", balsa_app.compose_email);
+	if(balsa_app.compose_email) {
+	    if(strncmp(balsa_app.compose_email, "mailto:", 7) == 0)
+	        sendmsg_window_process_url(balsa_app.compose_email+7, 
+		    	sendmsg_window_set_field, snd);
+	    else sendmsg_window_set_field(snd,"to", balsa_app.compose_email);
+	}
+	if (balsa_app.attach_file) {
+	    gchar *ptr = balsa_app.attach_file;
+	    gchar *delim;
+	    gchar *attachment;
+	    while ((delim = strchr (ptr, ','))) {
+		    *delim = '\0';
+		    attachment = g_strdup(ptr);
+		    add_attachment(GNOME_ICON_LIST(snd->attachments[1]), attachment);
+		    ptr = delim + 1;
+            }
+	    attachment = g_strdup(ptr);
+	    add_attachment(GNOME_ICON_LIST(snd->attachments[1]), attachment);
+	}
     } else
 	gtk_widget_show(window);
 
@@ -397,14 +366,16 @@ main(int argc, char *argv[])
 	gtk_idle_add((GtkFunction) open_mailboxes_idle_cb, names);
     }
 
+    signal( SIGPIPE, SIG_IGN );
+
     gdk_threads_enter();
     gtk_main();
     gdk_threads_leave();
-
+    
     gdk_colormap_unref(balsa_app.colormap);
 
 #ifdef BALSA_USE_THREADS
-    threads_init(FALSE);
+    threads_destroy();
 #endif
 
     return 0;
@@ -429,80 +400,59 @@ speed gain is negligible or negative: the number of called system
 functions in present case is constant and equal to 1; the previous
 version called system function either once or twice per directory. */
 static gboolean
-close_all_mailboxes(GNode * node, gpointer data)
+destroy_mbnode(GNode * node, gpointer data)
 {
-    MailboxNode *mbnode = (MailboxNode *) node->data;
+    BalsaMailboxNode *mbnode = (BalsaMailboxNode *) node->data;
 
-    if (mbnode) {
-	if (mbnode->IsDir) {
-	    gchar *tmpfile = g_strdup_printf("%s/.expanded", mbnode->name);
-	    if (mbnode->expanded)
-		creat(tmpfile, S_IRUSR | S_IWUSR);
-	    else
-		unlink(tmpfile);
-	    g_free(tmpfile);
-	}
-
-	force_close_mailbox(mbnode->mailbox);
+    if(mbnode == NULL) /* true for root node only */
+	return FALSE;
+    
+    if (!mbnode->mailbox) {
+	gchar *tmpfile = g_strdup_printf("%s/.expanded", mbnode->name);
+	if (mbnode->expanded)
+	    close(creat(tmpfile, S_IRUSR | S_IWUSR));
+	else
+	    unlink(tmpfile);
+	g_free(tmpfile);
     }
+    gtk_object_destroy(GTK_OBJECT(mbnode));
     return FALSE;
 }
 
-void
-balsa_exit(void)
+static void
+balsa_cleanup(void)
 {
-    g_node_traverse(balsa_app.mailbox_nodes,
-		    G_LEVEL_ORDER,
-		    G_TRAVERSE_ALL, 10, close_all_mailboxes, NULL);
-
     if (balsa_app.empty_trash_on_exit)
 	empty_trash();
 
-    force_close_mailbox(balsa_app.inbox);
-    force_close_mailbox(balsa_app.outbox);
-    force_close_mailbox(balsa_app.sentbox);
-    force_close_mailbox(balsa_app.draftbox);
-    force_close_mailbox(balsa_app.trash);
-
     config_save();
 
+    /* FIXME: stop switching notebook pages in a more elegant way.
+       Probably, the cleanest solution is to call enable_menus_xxx
+       functions from an idle function connected to balsa_message_set. */
+    gtk_signal_disconnect_by_data(GTK_OBJECT(balsa_app.notebook), NULL);
+
+    /* close all mailboxes */
+    g_node_traverse(balsa_app.mailbox_nodes,
+		    G_LEVEL_ORDER,
+		    G_TRAVERSE_ALL, 10, destroy_mbnode, NULL);
+    gtk_object_unref(GTK_OBJECT(balsa_app.inbox));
+    gtk_object_unref(GTK_OBJECT(balsa_app.outbox));
+    gtk_object_unref(GTK_OBJECT(balsa_app.sentbox));
+    gtk_object_unref(GTK_OBJECT(balsa_app.draftbox));
+    gtk_object_unref(GTK_OBJECT(balsa_app.trash));
+    g_node_destroy(balsa_app.mailbox_nodes);
+    balsa_app.mailbox_nodes = NULL;
     gnome_sound_shutdown();
-    gtk_main_quit();
+    libbalsa_imap_close_all_connections();
+    if(balsa_app.debug) g_print("Finished cleaning up.\n");
 }
 
-/* balsa_window_destroy
-   It may be called from balsa_window_destroy or balsa_exit; this is why
-   it should not make assumptions about the presence of the like
-   the notebook and so on.
-*/
-static void
-empty_trash(void)
-{
-    BalsaIndexPage *page;
-    GList *message;
-
-    libbalsa_mailbox_open(balsa_app.trash, FALSE);
-
-    message = balsa_app.trash->message_list;
-
-    while (message) {
-	libbalsa_message_delete(message->data);
-	message = message->next;
-    }
-    libbalsa_mailbox_commit_changes(balsa_app.trash);
-
-    libbalsa_mailbox_close(balsa_app.trash);
-
-    if (balsa_app.notebook &&
-	(page = balsa_find_notebook_page(balsa_app.trash)))
-	    balsa_index_page_reset(page);
-}
-
-
 static gint
 balsa_kill_session(GnomeClient * client, gpointer client_data)
 {
-    balsa_exit();
+    gtk_main_quit(); /* FIXME: this won't save composed messages; 
+			but it never did. */
     return TRUE;
 }
 
