@@ -1,6 +1,6 @@
 /* -*-mode:c; c-style:k&r; c-basic-offset:4; -*- */
 /* Balsa E-Mail Client
- * Copyright (C) 1997-2000 Stuart Parmenter and others,
+ * Copyright (C) 1997-2001 Stuart Parmenter and others,
  *                         See the file AUTHORS for a list.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -21,34 +21,39 @@
 
 #include "config.h"
 
-#include "libbalsa.h"
-
 #include <gnome.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <string.h>
 #include "balsa-app.h"
 #include "save-restore.h"
 #include "quote-color.h"
+#include "toolbar-prefs.h"
 
 #define BALSA_CONFIG_PREFIX "balsa/"
+#define FOLDER_SECTION_PREFIX "folder-"
 #define MAILBOX_SECTION_PREFIX "mailbox-"
 #define ADDRESS_BOOK_SECTION_PREFIX "address-book-"
+#define IDENTITY_SECTION_PREFIX "identity-"
 
-static gint config_mailboxes_init(void);
+static gint config_section_init(const char* section_prefix, 
+				gint (*cb)(const char*));
 static gint config_global_load(void);
+static gint config_folder_init(const gchar * prefix);
 static gint config_mailbox_init(const gchar * prefix);
 static gchar *config_get_unused_section(const gchar * prefix);
 
 static gchar **mailbox_list_to_vector(GList * mailbox_list);
 static void save_color(gchar * key, GdkColor * color);
 static void load_color(gchar * key, GdkColor * color);
+static void save_mru(GList *mru);
+static void load_mru(GList **mru);
 
 static void config_address_books_load(void);
 static void config_address_books_save(void);
-
 static void config_identities_load(void);
-static void config_identities_save(void);
+
+#define folder_section_path(mn) \
+    BALSA_MAILBOX_NODE(mn)->config_prefix ? \
+    g_strdup(BALSA_MAILBOX_NODE(mn)->config_prefix) : \
+    config_get_unused_section(FOLDER_SECTION_PREFIX)
 
 #define mailbox_section_path(mbox) \
     LIBBALSA_MAILBOX(mbox)->config_prefix ? \
@@ -62,8 +67,12 @@ static void config_identities_save(void);
 
 gint config_load(void)
 {
-    config_mailboxes_init();
-    config_global_load();
+    if(!config_global_load())  /* initializes balsa_app.mailbox_nodes */
+	return FALSE;          /* needed in the next step             */
+				
+    config_section_init(MAILBOX_SECTION_PREFIX, config_mailbox_init);
+    config_section_init(FOLDER_SECTION_PREFIX,  config_folder_init);
+
     return TRUE;
 }
 
@@ -75,6 +84,70 @@ d_get_gint(const gchar * key, gint def_val)
     return def ? def_val : res;
 }
 
+static void
+free_toolbars(void)
+{
+    int i, j;
+    
+    if(!balsa_app.toolbars == 0)
+	return;
+    
+    for(i=0; i<balsa_app.toolbar_count; i++) {
+	for(j=0; balsa_app.toolbars[i][j]; j++)
+	    g_free(balsa_app.toolbars[i][j]);
+	free((char *)balsa_app.toolbars[i]);
+    }
+    free((void *)balsa_app.toolbars);
+    free((void *)balsa_app.toolbar_ids);
+}
+
+/* load_toolbars:
+   loads customized toolbars for main/message preview and compose windows.
+*/
+static void
+load_toolbars(void)
+{
+    gint i, j, items;
+    char tmpkey[256];
+
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Toolbars/");
+    
+    balsa_app.toolbar_wrap_button_text=d_get_gint("WrapButtonText", 1);
+    
+    free_toolbars();
+    
+    balsa_app.toolbar_ids=
+	(int *)g_malloc(sizeof(int)*MAXTOOLBARS);
+    balsa_app.toolbars=	(char ***)g_malloc(sizeof(char *)*MAXTOOLBARS);
+    
+    balsa_app.toolbar_count=d_get_gint("ToolbarCount", 0);
+    if(balsa_app.toolbar_count)	{
+	for(i=0; i<balsa_app.toolbar_count; i++) {
+	    balsa_app.toolbars[i]=
+		(char **)g_malloc(sizeof(char *)*MAXTOOLBARITEMS);
+			
+	    sprintf(tmpkey, "Toolbar%dID", i);
+	    balsa_app.toolbar_ids[i] = d_get_gint(tmpkey, -1);
+
+	    if(balsa_app.toolbar_ids[i] == -1) {
+		balsa_app.toolbars[i][0]=NULL;
+		continue;
+	    }
+
+	    sprintf(tmpkey, "Toolbar%dItemCount", i);
+	    items=d_get_gint(tmpkey, 0);
+	    
+	    for(j=0; j<items; j++) {
+		sprintf(tmpkey, "Toolbar%dItem%d", i, j);
+		balsa_app.toolbars[i][j]=  gnome_config_get_string(tmpkey);
+	    }
+	    balsa_app.toolbars[i][j]=NULL;
+	}
+	
+    }
+    gnome_config_pop_prefix();
+}
+
 /* config_mailbox_set_as_special:
    allows to set given mailboxe as one of the special mailboxes
    PS: I am not sure if I should add outbox to the list.
@@ -87,7 +160,6 @@ void
 config_mailbox_set_as_special(LibBalsaMailbox * mailbox, specialType which)
 {
     LibBalsaMailbox **special;
-    GNode *node;
 
     g_return_if_fail(mailbox != NULL);
 
@@ -108,19 +180,16 @@ config_mailbox_set_as_special(LibBalsaMailbox * mailbox, specialType which)
 	return;
     }
     if (*special) {
+	g_free((*special)->config_prefix); 
+	(*special)->config_prefix = NULL;
 	config_mailbox_add(*special, NULL);
-	node = g_node_new(mailbox_node_new(
-					   (*special)->name, *special,
-					   (*special)->is_directory));
-	g_node_append(balsa_app.mailbox_nodes, node);
+	gtk_object_unref(GTK_OBJECT(*special));
     }
     config_mailbox_delete(mailbox);
     config_mailbox_add(mailbox, specialNames[which]);
 
-    node = find_gnode_in_mbox_list(balsa_app.mailbox_nodes, mailbox);
-    g_node_unlink(node);
-
     *special = mailbox;
+    gtk_object_ref(GTK_OBJECT(mailbox));
 }
 
 void
@@ -162,14 +231,40 @@ gint config_mailbox_add(LibBalsaMailbox * mailbox, const char *key_arg)
 	tmp = g_strdup_printf(BALSA_CONFIG_PREFIX MAILBOX_SECTION_PREFIX
 			      "%s/", key_arg);
 
+    gnome_config_push_prefix(tmp);
     libbalsa_mailbox_save_config(mailbox, tmp);
+    gnome_config_pop_prefix();
     g_free(tmp);
 
     gnome_config_sync();
     return TRUE;
 }				/* config_mailbox_add */
 
-/* removes from the ocnfiguration only */
+/* config_folder_add:
+   adds the specifed folder to the configuration. If the folder does
+   not have the unique pkey assigned, find one.
+*/
+gint config_folder_add(BalsaMailboxNode * mbnode, const char *key_arg)
+{
+    gchar *tmp;
+    g_return_val_if_fail(mbnode, FALSE);
+
+    if (key_arg == NULL)
+	tmp = folder_section_path(mbnode);
+    else
+	tmp = g_strdup_printf(BALSA_CONFIG_PREFIX FOLDER_SECTION_PREFIX
+			      "%s/", key_arg);
+
+    gnome_config_push_prefix(tmp);
+    balsa_mailbox_node_save_config(mbnode, tmp);
+    gnome_config_pop_prefix();
+    g_free(tmp);
+
+    gnome_config_sync();
+    return TRUE;
+}				/* config_mailbox_add */
+
+/* removes from the configuration only */
 gint config_mailbox_delete(const LibBalsaMailbox * mailbox)
 {
     gchar *tmp;			/* the key in the mailbox section name */
@@ -182,6 +277,19 @@ gint config_mailbox_delete(const LibBalsaMailbox * mailbox)
     return res;
 }				/* config_mailbox_delete */
 
+gint
+config_folder_delete(const BalsaMailboxNode * mbnode)
+{
+    gchar *tmp;			/* the key in the mailbox section name */
+    gint res;
+    tmp = folder_section_path(mbnode);
+    res = gnome_config_has_section(tmp);
+    gnome_config_clean_section(tmp);
+    gnome_config_sync();
+    g_free(tmp);
+    return res;
+}				/* config_folder_delete */
+
 /* Update the configuration information for the specified mailbox. */
 gint config_mailbox_update(LibBalsaMailbox * mailbox)
 {
@@ -197,27 +305,43 @@ gint config_mailbox_update(LibBalsaMailbox * mailbox)
     return res;
 }				/* config_mailbox_update */
 
+/* Update the configuration information for the specified folder. */
+gint config_folder_update(BalsaMailboxNode * mbnode)
+{
+    gchar *key;			/* the key in the mailbox section name */
+    gint res;
+
+    key = folder_section_path(mbnode);
+    res = gnome_config_has_section(key);
+    gnome_config_push_prefix(key);
+    balsa_mailbox_node_save_config(mbnode, key);
+    gnome_config_pop_prefix();
+    gnome_config_sync();
+    return res;
+}				/* config_folder_update */
+
 /* This function initializes all the mailboxes internally, going through
    the list of all the mailboxes in the configuration file one by one. */
+
 static gint
-config_mailboxes_init(void)
+config_section_init(const char* section_prefix, gint (*cb)(const char*))
 {
     void *iterator;
     gchar *key, *val, *tmp;
-    int pref_len = strlen(MAILBOX_SECTION_PREFIX);
+    int pref_len = strlen(section_prefix);
 
     iterator = gnome_config_init_iterator_sections(BALSA_CONFIG_PREFIX);
     while ((iterator = gnome_config_iterator_next(iterator, &key, &val))) {
-	if (strncmp(key, MAILBOX_SECTION_PREFIX, pref_len) == 0) {
+	if (strncmp(key, section_prefix, pref_len) == 0) {
 	    tmp = g_strconcat(BALSA_CONFIG_PREFIX, key, "/", NULL);
-	    config_mailbox_init(tmp);
+	    cb(tmp);
 	    g_free(tmp);
 	}
 	g_free(key);
 	g_free(val);
     }
-    return TRUE;		/* hm... check_basic_mailboxes? */
-}				/* config_mailboxes_init */
+    return TRUE;
+}				/* config_section_init */
 
 /* Initialize the specified mailbox, creating the internal data
    structures which represent the mailbox. */
@@ -228,7 +352,6 @@ config_mailbox_init(const gchar * prefix)
     const gchar *key =
 	prefix + strlen(BALSA_CONFIG_PREFIX MAILBOX_SECTION_PREFIX);
     GNode *node;
-    gboolean is_mh;
 
     g_return_val_if_fail(prefix != NULL, FALSE);
 
@@ -245,40 +368,61 @@ config_mailbox_init(const gchar * prefix)
 
     if (LIBBALSA_IS_MAILBOX_POP3(mailbox)) {
 	balsa_app.inbox_input =
-	    g_list_append(balsa_app.inbox_input, mailbox);
-    } else if (strcmp("Inbox/", key) == 0)
-	balsa_app.inbox = mailbox;
-    else if (strcmp("Outbox/", key) == 0)
-	balsa_app.outbox = mailbox;
-    else if (strcmp("Sentbox/", key) == 0)
-	balsa_app.sentbox = mailbox;
-    else if (strcmp("Draftbox/", key) == 0)
-	balsa_app.draftbox = mailbox;
-    else if (strcmp("Trash/", key) == 0)
-	balsa_app.trash = mailbox;
-    else {
-	is_mh = LIBBALSA_IS_MAILBOX_LOCAL(mailbox)
-	    && LIBBALSA_MAILBOX_LOCAL(mailbox)->type ==
-	    LIBBALSA_MAILBOX_LOCAL_MH;
+	    g_list_append(balsa_app.inbox_input, 
+			  balsa_mailbox_node_new_from_mailbox(mailbox));
+    } else {
+        gboolean special = TRUE;
 
-	node = g_node_new(mailbox_node_new(g_strdup(mailbox->name),
-					   mailbox, is_mh));
+	node = g_node_new(balsa_mailbox_node_new_from_mailbox(mailbox));
 	g_node_append(balsa_app.mailbox_nodes, node);
+
+	if (strcmp("Inbox/", key) == 0)
+	    balsa_app.inbox = mailbox;
+	else if (strcmp("Outbox/", key) == 0)
+	    balsa_app.outbox = mailbox;
+	else if (strcmp("Sentbox/", key) == 0)
+	    balsa_app.sentbox = mailbox;
+	else if (strcmp("Draftbox/", key) == 0)
+	    balsa_app.draftbox = mailbox;
+	else if (strcmp("Trash/", key) == 0)
+	    balsa_app.trash = mailbox;
+	else
+            special = FALSE;
+
+	if (special)
+            gtk_object_ref(GTK_OBJECT(mailbox));
     }
     return TRUE;
 }				/* config_mailbox_init */
 
 
+/* Initialize the specified folder, creating the internal data
+   structures which represent the folder. */
+static gint
+config_folder_init(const gchar * prefix)
+{
+    BalsaMailboxNode *folder;
+
+    g_return_val_if_fail(prefix != NULL, FALSE);
+
+    if( (folder = balsa_mailbox_node_new_from_config(prefix)) )
+	g_node_append(balsa_app.mailbox_nodes, g_node_new(folder));
+
+    return folder != NULL;
+}				/* config_folder_init */
+
 /* Load Balsa's global settings */
 static gint
 config_global_load(void)
 {
     gchar **open_mailbox_vector;
     gint open_mailbox_count;
-
+#if ENABLE_ESMTP
+    gboolean def_used;
+#endif
     config_address_books_load();
     config_identities_load();
-
+    
     /* Section for the balsa_information() settings... */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "InformationMessages/");
 
@@ -290,6 +434,8 @@ config_global_load(void)
 	d_get_gint("ShowErrorMessages", BALSA_INFORMATION_SHOW_DIALOG);
     balsa_app.debug_message =
 	d_get_gint("ShowDebugMessages", BALSA_INFORMATION_SHOW_NONE);
+    balsa_app.fatal_message = 
+	d_get_gint("ShowFatalMessages", BALSA_INFORMATION_SHOW_DIALOG);
 
     gnome_config_pop_prefix();
 
@@ -316,11 +462,16 @@ config_global_load(void)
 	d_get_gint("IndexSubjectWidth", SUBJECT_DEFAULT_WIDTH);
     balsa_app.index_date_width =
 	d_get_gint("IndexDateWidth", DATE_DEFAULT_WIDTH);
+    balsa_app.index_size_width =
+	d_get_gint("IndexSizeWidth", SIZE_DEFAULT_WIDTH);
 
     /* ... window sizes */
     balsa_app.mw_width = gnome_config_get_int("MainWindowWidth=640");
     balsa_app.mw_height = gnome_config_get_int("MainWindowHeight=480");
     balsa_app.mblist_width = gnome_config_get_int("MailboxListWidth=100");
+    /* sendmsg window sizes */
+    balsa_app.sw_width = gnome_config_get_int("SendMsgWindowWidth=640");
+    balsa_app.sw_height = gnome_config_get_int("SendMsgWindowHeight=480");
     /* FIXME: PKGW: why comment this out? Breaks my Transfer context menu. */
     if (balsa_app.mblist_width < 100)
 	balsa_app.mblist_width = 170;
@@ -347,16 +498,32 @@ config_global_load(void)
     balsa_app.selected_headers =
 	gnome_config_get_string("SelectedHeaders=" DEFAULT_SELECTED_HDRS);
     g_strdown(balsa_app.selected_headers);
+    balsa_app.threading_type = d_get_gint("ThreadingType", 
+					  BALSA_INDEX_THREADING_JWZ);
 
     /* ... Quote colouring */
     g_free(balsa_app.quote_regex);
     balsa_app.quote_regex =
 	gnome_config_get_string("QuoteRegex=" DEFAULT_QUOTE_REGEX);
-    load_color("QuotedColorStart=" DEFAULT_QUOTED_COLOR,
-	       &balsa_app.quoted_color[0]);
-    load_color("QuotedColorEnd=" DEFAULT_QUOTED_COLOR,
-	       &balsa_app.quoted_color[MAX_QUOTED_COLOR - 1]);
-    make_gradient(balsa_app.quoted_color, 0, MAX_QUOTED_COLOR - 1);
+
+    {
+	int i;
+#if MAX_QUOTED_COLOR != 6
+#warning 'default_quoted_color' array needs to be updated
+#endif
+	gchar *default_quoted_color[6] = {
+	    "rgb:0000/8000/8000", "rgb:8000/0000/0000", "rgb:0000/8000/0000",
+	    "rgb:0000/0000/8000", "rgb:8000/8000/0000", "rgb:8000/0000/8000"};
+	for(i=0;i<MAX_QUOTED_COLOR;i++) {
+	    gchar *text = g_strdup_printf("QuotedColor%d=%s", i, i<6 ?
+			  default_quoted_color[i] : DEFAULT_QUOTED_COLOR);
+	    load_color(text, &balsa_app.quoted_color[i]);
+	    g_free(text);
+	}
+    }
+
+    /* URL coloring */
+    load_color("UrlColor=" DEFAULT_URL_COLOR, &balsa_app.url_color);
 
     /* ... font used to display messages */
     g_free(balsa_app.message_font);
@@ -379,23 +546,30 @@ config_global_load(void)
     balsa_app.show_mblist = gnome_config_get_bool("ShowMailboxList=true");
     balsa_app.show_notebook_tabs = gnome_config_get_bool("ShowTabs=false");
 
+    /* ... alternative layout of main window */
+    balsa_app.alternative_layout = gnome_config_get_bool("AlternativeLayout=false");
+    balsa_app.view_message_on_open = gnome_config_get_bool("ViewMessageOnOpen=true");
+    balsa_app.line_length = gnome_config_get_bool("MsgSizeAsLines=true");
+    balsa_app.pgdownmod = gnome_config_get_bool("PageDownMod=false");
+    balsa_app.pgdown_percent = gnome_config_get_int("PageDownPercent=50");
+    if (balsa_app.pgdown_percent < 10)
+	balsa_app.pgdown_percent = 10;
+
     /* ... style */
     balsa_app.toolbar_style = d_get_gint("ToolbarStyle", GTK_TOOLBAR_BOTH);
     /* ... Progress Window Dialog */
     balsa_app.pwindow_option = d_get_gint("ProgressWindow", WHILERETR);
+    balsa_app.drag_default_is_move = d_get_gint("DragDefaultIsMove", 0);
 
     gnome_config_pop_prefix();
 
     /* Printing options ... */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Printing/");
 
-    g_free(balsa_app.PrintCommand.PrintCommand);
-    balsa_app.PrintCommand.PrintCommand =
-	gnome_config_get_string("PrintCommand=a2ps -d -q %s");
-    balsa_app.PrintCommand.linesize =
-	d_get_gint("PrintLinesize", DEFAULT_LINESIZE);
-    balsa_app.PrintCommand.breakline =
-	gnome_config_get_bool("PrintBreakline=false");
+    /* ... Printing */
+    g_free(balsa_app.paper_size);
+    balsa_app.paper_size =
+	gnome_config_get_string("PaperSize=" DEFAULT_PAPER_SIZE);
 
     gnome_config_pop_prefix();
 
@@ -429,6 +603,10 @@ config_global_load(void)
     /* Maibox checking options ... */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "MailboxChecking/");
 
+    balsa_app.notify_new_mail_dialog =
+	d_get_gint("NewMailNotificationDialog", 0);
+    balsa_app.notify_new_mail_sound =
+	d_get_gint("NewMailNotificationSound", 1);
     balsa_app.check_mail_upon_startup =
 	gnome_config_get_bool("OnStartup=false");
     balsa_app.check_mail_auto = gnome_config_get_bool("Auto=false");
@@ -437,26 +615,59 @@ config_global_load(void)
 	balsa_app.check_mail_timer = 10;
     if (balsa_app.check_mail_auto)
 	update_timer(TRUE, balsa_app.check_mail_timer);
+    balsa_app.check_imap=d_get_gint("CheckIMAP", 1);
+    balsa_app.check_imap_inbox=d_get_gint("CheckIMAPInbox", 0);
+    balsa_app.quiet_background_check=d_get_gint("QuietBackgroundCheck", 0);
+    gnome_config_pop_prefix();
+
+#ifdef BALSA_MDN_REPLY
+    /* how to react if a message with MDN request is displayed */
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "MDNReply/");
+    balsa_app.mdn_reply_clean = gnome_config_get_int("Clean=1");
+    balsa_app.mdn_reply_notclean = gnome_config_get_int("Suspicious=0");
 
     gnome_config_pop_prefix();
+#endif
 
     /* Sending options ... */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Sending/");
 
+#if ENABLE_ESMTP
     /* ... SMTP server */
-    balsa_app.smtp_server = gnome_config_get_string("SMTPServer");
-    balsa_app.smtp = gnome_config_get_bool("SMTP=false");
-
+    balsa_app.smtp_server =
+	gnome_config_get_string_with_default("ESMTPServer=localhost:25", 
+					     &def_used);
+    if(def_used) {
+	/* we need to check for old format, 1.1.4-compatible settings and
+	   convert them if needed.
+	*/
+	gchar* old_server = gnome_config_get_string("SMTPServer");
+	if(old_server) {
+	    int port = gnome_config_get_int("SMTPPort=25");
+	    g_free(balsa_app.smtp_server);
+	    balsa_app.smtp_server = g_strdup_printf("%s:%d",
+						    old_server, port);
+	    g_free(old_server);
+	g_warning("Converted old SMTP server config to ESMTP format. Verify the correctness.");
+	}
+    }
+    balsa_app.smtp_user = gnome_config_get_string("ESMTPUser");
+    balsa_app.smtp_passphrase = gnome_config_get_string("ESMTPPassphrase");
+    balsa_app.smtp_tls_mode = gnome_config_get_int("ESMTPTLSMode=0");
+#if HAVE_SMTP_TLS_CLIENT_CERTIFICATE
+    balsa_app.smtp_certificate_passphrase = gnome_config_get_string("ESMTPCertificatePassphrase");
+#endif
+#endif
     /* ... outgoing mail */
     balsa_app.encoding_style = gnome_config_get_int("EncodingStyle=2");
-    g_free(balsa_app.charset);
-    balsa_app.charset =
-	gnome_config_get_string("Charset=" DEFAULT_CHARSET);
-    libbalsa_set_charset(balsa_app.charset);
     balsa_app.wordwrap = gnome_config_get_bool("WordWrap=true");
     balsa_app.wraplength = gnome_config_get_int("WrapLength=75");
     if (balsa_app.wraplength < 40)
 	balsa_app.wraplength = 40;
+    balsa_app.reply_strip_html = 
+	gnome_config_get_bool("StripHtmlInReply=true");
+
+	balsa_app.always_queue_sent_mail = d_get_gint("AlwaysQueueSentMail", 0);
 
     gnome_config_pop_prefix();
 
@@ -468,6 +679,8 @@ config_global_load(void)
     g_free(balsa_app.compose_headers);
     balsa_app.compose_headers =
 	gnome_config_get_string("ComposeHeaders=to subject cc");
+    balsa_app.req_dispnotify = 
+	gnome_config_get_bool("RequestDispositionNotification=false");
 
     gnome_config_pop_prefix();
 
@@ -478,15 +691,26 @@ config_global_load(void)
     g_free(balsa_app.local_mail_directory);
     balsa_app.local_mail_directory = gnome_config_get_string("MailDir");
 
+    if(!balsa_app.local_mail_directory) {
+	gnome_config_pop_prefix();
+	return FALSE;
+    }
+    balsa_app.mailbox_nodes = g_node_new(balsa_mailbox_node_new_from_dir(
+        balsa_app.local_mail_directory));
+
     /* debugging enabled */
     balsa_app.debug = gnome_config_get_bool("Debug=false");
 
+    balsa_app.close_mailbox_auto = gnome_config_get_bool("AutoCloseMailbox=true");
+    balsa_app.close_mailbox_timeout = gnome_config_get_int("AutoCloseMailboxTimeout=10");
+
     balsa_app.remember_open_mboxes =
 	gnome_config_get_bool("RememberOpenMailboxes=false");
     gnome_config_get_vector("OpenMailboxes", &open_mailbox_count,
 			    &open_mailbox_vector);
     if (balsa_app.remember_open_mboxes && open_mailbox_count > 0) {
 	/* FIXME: Open the mailboxes.... */
+	printf("Opening %d mailboxes on startup.\n", open_mailbox_count);
 	gtk_idle_add((GtkFunction) open_mailboxes_idle_cb,
 		     open_mailbox_vector);
     } else
@@ -494,16 +718,36 @@ config_global_load(void)
 
     balsa_app.empty_trash_on_exit =
 	gnome_config_get_bool("EmptyTrash=false");
-    balsa_app.ab_dist_list_mode =
-	gnome_config_get_bool("AddressBookDistMode=false");
+
+    /* This setting is now per address book */
+    gnome_config_clean_key("AddressBookDistMode");
 
     gnome_config_pop_prefix();
+    
+    /* Toolbars */
+    load_toolbars();
+
+    /* Last used paths options ... */
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Paths/");
+    g_free(balsa_app.attach_dir);
+    balsa_app.attach_dir = gnome_config_get_string("AttachDir");
+    g_free(balsa_app.save_dir);
+    balsa_app.save_dir = gnome_config_get_string("SavePartDir");
+    gnome_config_pop_prefix();
+
+	/* Folder MRU */
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "FolderMRU/");
+	load_mru(&balsa_app.folder_mru);
+    gnome_config_pop_prefix();
+
     return TRUE;
 }				/* config_global_load */
 
 gint config_save(void)
 {
     gchar **open_mailboxes_vector;
+    gint i, j;
+	char tmpkey[32];
 
     config_address_books_save();
     config_identities_save();
@@ -515,9 +759,11 @@ gint config_save(void)
     gnome_config_set_int("ShowWarningMessages", balsa_app.warning_message);
     gnome_config_set_int("ShowErrorMessages", balsa_app.error_message);
     gnome_config_set_int("ShowDebugMessages", balsa_app.debug_message);
+    gnome_config_set_int("ShowFatalMessages", balsa_app.fatal_message);
     gnome_config_pop_prefix();
 
-/* Section for geometry ... *//* FIXME: Saving window sizes is the WM's job?? */
+    /* Section for geometry ... */
+    /* FIXME: Saving window sizes is the WM's job?? */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Geometry/");
 
     /* ... column width settings */
@@ -535,9 +781,12 @@ gint config_save(void)
     gnome_config_set_int("IndexSubjectWidth",
 			 balsa_app.index_subject_width);
     gnome_config_set_int("IndexDateWidth", balsa_app.index_date_width);
+    gnome_config_set_int("IndexSizeWidth", balsa_app.index_size_width);
     gnome_config_set_int("MainWindowWidth", balsa_app.mw_width);
     gnome_config_set_int("MainWindowHeight", balsa_app.mw_height);
     gnome_config_set_int("MailboxListWidth", balsa_app.mblist_width);
+    gnome_config_set_int("SendMsgWindowWidth", balsa_app.sw_width);
+    gnome_config_set_int("SendMsgWindowHeight", balsa_app.sw_height);
     gnome_config_set_int("NotebookHeight", balsa_app.notebook_height);
 
     gnome_config_pop_prefix();
@@ -548,13 +797,19 @@ gint config_save(void)
     gnome_config_set_string("DateFormat", balsa_app.date_string);
     gnome_config_set_int("ShownHeaders", balsa_app.shown_headers);
     gnome_config_set_string("SelectedHeaders", balsa_app.selected_headers);
+    gnome_config_set_int("ThreadingType", balsa_app.threading_type);
     gnome_config_set_string("QuoteRegex", balsa_app.quote_regex);
     gnome_config_set_string("MessageFont", balsa_app.message_font);
     gnome_config_set_string("SubjectFont", balsa_app.subject_font);
     gnome_config_set_bool("WordWrap", balsa_app.browse_wrap);
-    save_color("QuotedColorStart", &balsa_app.quoted_color[0]);
-    save_color("QuotedColorEnd",
-	       &balsa_app.quoted_color[MAX_QUOTED_COLOR - 1]);
+
+    for(i=0;i<MAX_QUOTED_COLOR;i++) {
+	gchar *text = g_strdup_printf("QuotedColor%d", i);
+	save_color(text, &balsa_app.quoted_color[i]);
+	g_free(text);
+    }
+
+    save_color("UrlColor", &balsa_app.url_color);
 
     gnome_config_pop_prefix();
 
@@ -566,18 +821,18 @@ gint config_save(void)
     gnome_config_set_bool("ShowTabs", balsa_app.show_notebook_tabs);
     gnome_config_set_int("ToolbarStyle", balsa_app.toolbar_style);
     gnome_config_set_int("ProgressWindow", balsa_app.pwindow_option);
+    gnome_config_set_bool("AlternativeLayout", balsa_app.alternative_layout);
+    gnome_config_set_bool("ViewMessageOnOpen", balsa_app.view_message_on_open);
+    gnome_config_set_bool("MsgSizeAsLines", balsa_app.line_length);
+    gnome_config_set_bool("PageDownMod", balsa_app.pgdownmod);
+    gnome_config_set_int("PageDownPercent", balsa_app.pgdown_percent);
+    gnome_config_set_int("DragDefaultIsMove", balsa_app.drag_default_is_move);
 
     gnome_config_pop_prefix();
 
     /* Printing options ... */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Printing/");
-
-    gnome_config_set_string("PrintCommand",
-			    balsa_app.PrintCommand.PrintCommand);
-    gnome_config_set_int("PrintLinesize", balsa_app.PrintCommand.linesize);
-    gnome_config_set_bool("PrintBreakline",
-			  balsa_app.PrintCommand.breakline);
-
+    gnome_config_set_string("PaperSize",balsa_app.paper_size);
     gnome_config_pop_prefix();
 
     /* Spelling options ... */
@@ -603,28 +858,53 @@ gint config_save(void)
     /* Maibox checking options ... */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "MailboxChecking/");
 
+	gnome_config_set_int("NewMailNotificationDialog",
+				balsa_app.notify_new_mail_dialog);
+	gnome_config_set_int("NewMailNotificationSound",
+				balsa_app.notify_new_mail_sound);
     gnome_config_set_bool("OnStartup", balsa_app.check_mail_upon_startup);
     gnome_config_set_bool("Auto", balsa_app.check_mail_auto);
     gnome_config_set_int("AutoDelay", balsa_app.check_mail_timer);
+    gnome_config_set_int("CheckIMAP", balsa_app.check_imap);
+    gnome_config_set_int("CheckIMAPInbox", balsa_app.check_imap_inbox);
+    gnome_config_set_int("QuietBackgroundCheck",
+			 balsa_app.quiet_background_check);
 
     gnome_config_pop_prefix();
 
+#ifdef BALSA_MDN_REPLY
+    /* how to react if a message with MDN request is displayed */
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "MDNReply/");
+    gnome_config_set_int("Clean", balsa_app.mdn_reply_clean);
+    gnome_config_set_int("Suspicious", balsa_app.mdn_reply_notclean);
+
+    gnome_config_pop_prefix();
+#endif
+
     /* Sending options ... */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Sending/");
-
-    gnome_config_set_bool("SMTP", balsa_app.smtp);
-    gnome_config_set_string("SMTPServer", balsa_app.smtp_server);
-    gnome_config_set_int("EncodingStyle", balsa_app.encoding_style);
-    gnome_config_set_string("Charset", balsa_app.charset);
+#if ENABLE_ESMTP
+    gnome_config_set_string("ESMTPServer", balsa_app.smtp_server);
+    gnome_config_set_string("ESMTPUser", balsa_app.smtp_user);
+    gnome_config_set_string("ESMTPPassphrase", balsa_app.smtp_passphrase);
+    gnome_config_set_int("ESMTPTLSMode", balsa_app.smtp_tls_mode);
+#if HAVE_SMTP_TLS_CLIENT_CERTIFICATE
+    gnome_config_set_string("ESMTPCertificatePassphrase", balsa_app.smtp_certificate_passphrase);
+#endif 
+#endif 
+   gnome_config_set_int("EncodingStyle", balsa_app.encoding_style);
     gnome_config_set_bool("WordWrap", balsa_app.wordwrap);
     gnome_config_set_int("WrapLength", balsa_app.wraplength);
+    gnome_config_set_bool("StripHtmlInReply", balsa_app.reply_strip_html);
 
+	gnome_config_set_int("AlwaysQueueSentMail", balsa_app.always_queue_sent_mail);
     gnome_config_pop_prefix();
 
     /* Compose window ... */
     gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Compose/");
 
     gnome_config_set_string("ComposeHeaders", balsa_app.compose_headers);
+    gnome_config_set_bool("RequestDispositionNotification", balsa_app.req_dispnotify);
     gnome_config_set_string("QuoteString", balsa_app.quote_str);
 
     gnome_config_pop_prefix();
@@ -636,13 +916,15 @@ gint config_save(void)
 
     gnome_config_set_bool("Debug", balsa_app.debug);
 
+    gnome_config_set_bool("AutoCloseMailbox", balsa_app.close_mailbox_auto);
+    gnome_config_set_int("AutoCloseMailboxTimeout", balsa_app.close_mailbox_timeout);
+
     open_mailboxes_vector =
 	mailbox_list_to_vector(balsa_app.open_mailbox_list);
     gnome_config_set_vector("OpenMailboxes",
 			    g_list_length(balsa_app.open_mailbox_list),
 			    (const char **) open_mailboxes_vector);
     g_strfreev(open_mailboxes_vector);
-
     gnome_config_set_bool("RememberOpenMailboxes",
 			  balsa_app.remember_open_mboxes);
     gnome_config_set_bool("EmptyTrash", balsa_app.empty_trash_on_exit);
@@ -656,9 +938,44 @@ gint config_save(void)
 	gnome_config_clean_key("DefaultAddressBook");
     }
 
-    /* address book */
-    gnome_config_set_bool("AddressBookDistMode",
-			  balsa_app.ab_dist_list_mode);
+    gnome_config_pop_prefix();
+
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Toolbars/");
+
+	gnome_config_set_int("WrapButtonText", balsa_app.toolbar_wrap_button_text);
+
+	gnome_config_set_int("ToolbarCount", balsa_app.toolbar_count);
+
+	if(balsa_app.toolbar_count > 0)
+	{
+		for(i=0;i<balsa_app.toolbar_count;i++)
+		{
+			sprintf(tmpkey, "Toolbar%dID", i);
+			gnome_config_set_int(tmpkey, balsa_app.toolbar_ids[i]);
+
+			for(j=0;balsa_app.toolbars[i][j];j++)
+			{
+				sprintf(tmpkey, "Toolbar%dItem%d", i, j);
+				gnome_config_set_string(tmpkey, balsa_app.toolbars[i][j]);
+			}
+			sprintf(tmpkey, "Toolbar%dItemCount", i);
+			gnome_config_set_int(tmpkey, j);
+		}
+	}
+
+    gnome_config_pop_prefix();
+    
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "Paths/");
+    if(balsa_app.attach_dir)
+	gnome_config_set_string("AttachDir", balsa_app.attach_dir);
+    if(balsa_app.save_dir)
+	gnome_config_set_string("SavePartDir", balsa_app.save_dir);
+    gnome_config_pop_prefix();
+
+	
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "FolderMRU/");
+    save_mru(balsa_app.folder_mru);
+    gnome_config_pop_prefix();
 
     gnome_config_sync();
     return TRUE;
@@ -739,79 +1056,107 @@ static void
 config_address_books_save(void)
 {
     GList *list;
-    LibBalsaAddressBook *ab;
 
-    list = balsa_app.address_book_list;
-    while (list) {
-	ab = LIBBALSA_ADDRESS_BOOK(list->data);
-
-	config_address_book_save(ab);
-	list = g_list_next(list);
-
-    }
+    for (list=balsa_app.address_book_list; list; list=g_list_next(list))
+	config_address_book_save(LIBBALSA_ADDRESS_BOOK(list->data));
 }
 
+
 static void
-config_identities_load(void)
+config_identities_load()
 {
-    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "identity-default/");
+    LibBalsaIdentity* ident;
+    gchar *default_ident;
+    void *iterator;
+    gchar *key, *val, *tmp;
+    int pref_len = strlen(IDENTITY_SECTION_PREFIX);
 
-    if (balsa_app.address)
-	gtk_object_destroy(GTK_OBJECT(balsa_app.address));
-    balsa_app.address = libbalsa_address_new();
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "identity/");
+    default_ident = gnome_config_get_string("CurrentIdentity");
+    gnome_config_pop_prefix();
 
-    /* user's real name */
-    balsa_app.address->full_name = gnome_config_get_string("FullName");
-
-    /* user's email address */
-    balsa_app.address->address_list =
-	g_list_append(balsa_app.address->address_list,
-		      gnome_config_get_string("Address"));
-
-    /* users's replyto address */
-    balsa_app.replyto = gnome_config_get_string("ReplyTo");
-
-    /* users's domain */
-    balsa_app.domain = gnome_config_get_string("Domain");
-
-    /* bcc field for outgoing mails; optional */
-    balsa_app.bcc = gnome_config_get_string("Bcc");
-
-    /* signature file path */
-    balsa_app.signature_path = gnome_config_get_string("SignaturePath");
-    if (balsa_app.signature_path == NULL) {
-	balsa_app.signature_path =
-	    gnome_util_prepend_user_home(".signature");
+    iterator = gnome_config_init_iterator_sections(BALSA_CONFIG_PREFIX);
+    while ((iterator = gnome_config_iterator_next(iterator, &key, &val))) {
+	if (strncmp(key, IDENTITY_SECTION_PREFIX, pref_len) == 0) {
+	    tmp = g_strconcat(BALSA_CONFIG_PREFIX, key, "/", NULL);
+	    ident = libbalsa_identity_new_config(tmp, key+pref_len);
+	    balsa_app.identities = g_list_append(balsa_app.identities, ident);
+	    g_free(tmp);
+	    if(g_strcasecmp(default_ident, ident->identity_name) == 0)
+		balsa_app.current_ident = ident;
+	}
+	g_free(key);
+	g_free(val);
     }
 
-    balsa_app.sig_sending = gnome_config_get_bool("SigSending=true");
-    balsa_app.sig_whenreply = gnome_config_get_bool("SigReply=true");
-    balsa_app.sig_whenforward = gnome_config_get_bool("SigForward=true");
-    balsa_app.sig_separator = gnome_config_get_bool("SigSeparator=true");
+    if (g_list_length(balsa_app.identities) == 0)
+        balsa_app.identities = 
+	    g_list_append(balsa_app.identities,
+			  libbalsa_identity_new_config(BALSA_CONFIG_PREFIX
+						       "identity-default/",
+						       "default"));
+    
+    if(balsa_app.current_ident == NULL)
+	balsa_app.current_ident = 
+	    LIBBALSA_IDENTITY(balsa_app.identities->data);
 
-    gnome_config_pop_prefix();
+    g_free(default_ident);
 }
 
-static void
+/* config_identities_save:
+   saves the balsa_app.identites list.
+*/
+void
 config_identities_save(void)
 {
-    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "identity-default/");
+    LibBalsaIdentity* ident;
+    GList* list;
+    gchar** conf_vec, *prefix, *key, *val;
+    GList* old_identities = NULL;
+    gint i = 0;
+    void* iterator;
+    int pref_len = strlen(IDENTITY_SECTION_PREFIX);
 
-    gnome_config_set_string("FullName", balsa_app.address->full_name);
-    gnome_config_set_string("Address",
-			    balsa_app.address->address_list->data);
-    gnome_config_set_string("ReplyTo", balsa_app.replyto);
-    gnome_config_set_string("Domain", balsa_app.domain);
-    gnome_config_set_string("Bcc", balsa_app.bcc);
-    gnome_config_set_string("SignaturePath", balsa_app.signature_path);
+    conf_vec = g_malloc(sizeof(gchar*) * g_list_length(balsa_app.identities));
 
-    gnome_config_set_bool("SigSending", balsa_app.sig_sending);
-    gnome_config_set_bool("SigForward", balsa_app.sig_whenforward);
-    gnome_config_set_bool("SigReply", balsa_app.sig_whenreply);
-    gnome_config_set_bool("SigSeparator", balsa_app.sig_separator);
+    g_assert(conf_vec != NULL);
 
+    for (list = balsa_app.identities; list; list = g_list_next(list)) {
+        ident = LIBBALSA_IDENTITY(list->data);
+        conf_vec[i++] = ident->identity_name;
+    }
+    
+    gnome_config_push_prefix(BALSA_CONFIG_PREFIX "identity/");
+    gnome_config_set_vector("Identities", i, (const char**) conf_vec);
+    gnome_config_set_string("CurrentIdentity", 
+                            balsa_app.current_ident->identity_name);
     gnome_config_pop_prefix();
+    g_free(conf_vec);
 
+    /* clean removed sections */
+    iterator = gnome_config_init_iterator_sections(BALSA_CONFIG_PREFIX);
+    while ((iterator = gnome_config_iterator_next(iterator, &key, &val))) {
+	if (strncmp(key, IDENTITY_SECTION_PREFIX, pref_len) == 0) {
+	    prefix = g_strconcat(BALSA_CONFIG_PREFIX, key, "/", NULL);
+	    old_identities = g_list_prepend(old_identities, prefix);
+	}
+	g_free(key);
+	g_free(val);
+    }
+    for(list=old_identities; list; list = g_list_next(list)) {
+	gnome_config_clean_section(list->data);
+	g_free(list->data);
+    }
+    g_list_free(old_identities);
+
+    /* save current */
+    for (list = balsa_app.identities; list; list = g_list_next(list)) {
+	ident = LIBBALSA_IDENTITY(list->data);
+	prefix = g_strconcat(BALSA_CONFIG_PREFIX "identity-", 
+			     ident->identity_name, "/", NULL);
+        libbalsa_identity_save(ident, prefix);
+	g_free(prefix);
+    }
 }
 
 static gchar **
@@ -824,14 +1169,10 @@ mailbox_list_to_vector(GList * mailbox_list)
     i = g_list_length(mailbox_list) + 1;
     res = g_new0(gchar *, i);
 
-    i = 0;
-    list = mailbox_list;
-    while (list) {
-	res[i] = g_strdup(LIBBALSA_MAILBOX(list->data)->name);
-	i++;
-	list = g_list_next(list);
-    }
-    res[i] = NULL;
+    res[--i] = NULL;
+    for(list = mailbox_list; list; list = g_list_next(list))
+	res[--i] = g_strdup(LIBBALSA_MAILBOX(list->data)->name);
+
     return res;
 }
 
@@ -856,3 +1197,31 @@ load_color(gchar * key, GdkColor * color)
     gdk_color_parse(str, color);
     g_free(str);
 }
+
+static void
+load_mru(GList **mru)
+{
+    int count, i;
+    char tmpkey[32];
+    
+    count=d_get_gint("MRUCount", 0);
+    for(i=0;i<count;i++) {
+	sprintf(tmpkey, "MRU%d", i+1);
+	(*mru)=g_list_append((*mru), gnome_config_get_string(tmpkey));
+    }
+}
+
+static void
+save_mru(GList *mru)
+{
+    int i;
+    char tmpkey[32];
+    GList *ltmp;
+    
+    for(ltmp=g_list_first(balsa_app.folder_mru),i=0;
+	ltmp; ltmp=g_list_next(ltmp),i++) {
+	sprintf(tmpkey, "MRU%d", i+1);
+	gnome_config_set_string(tmpkey, (gchar *)(ltmp->data));
+    }
+    gnome_config_set_int("MRUCount", i);
+}
