@@ -30,6 +30,7 @@ extern "C" {
 	SEND_NORMAL,		/* initialized by Compose */
 	SEND_REPLY,		/* by Reply               */
 	SEND_REPLY_ALL,		/* by Reply All           */
+	SEND_REPLY_GROUP,       /* by Reply to Group      */
 	SEND_FORWARD,		/* by Forward             */
 	SEND_CONTINUE		/* by Continue postponed  */
     } SendType;
@@ -54,17 +55,31 @@ extern "C" {
 	const gchar *charset;
 	const gchar *locale;
 	GtkWidget *current_language_menu;
-
+	/* identity related data */
+	LibBalsaIdentity* ident;
 	/* widgets to be disabled when the address is incorrect */
-	GtkWidget *ready_widgets[2];
+	GtkWidget *ready_widgets[5];
 	GtkWidget *view_checkitems[VIEW_MENU_LENGTH];
 	GList *spell_check_disable_list;
+	gboolean update_config; /* is the window being set up or in normal  */
+	                        /* operation and user actions should update */
+	                        /* the config */
+	guint delete_sig_id;
+	gboolean modified;
     };
 
     BalsaSendmsg *sendmsg_window_new(GtkWidget *, LibBalsaMessage *,
 				     SendType);
     void sendmsg_window_set_field(BalsaSendmsg *bsmsg, const gchar* key,
 				  const gchar* val);
+
+    void add_attachment(GnomeIconList * iconlist, char *filename);
+
+    typedef void (*field_setter)(BalsaSendmsg *d, const gchar*, const gchar*);
+
+    void sendmsg_window_process_url(const char *url, field_setter func,
+				    void *data);
+
 #ifdef __cplusplus
 }
 #endif				/* __cplusplus */
